"""
Tests of neo.io.neuronexusio
"""

import unittest

from neo.io import NeuroNexusIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestNeuroNexusIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = NeuroNexusIO
    entities_to_download = ["neuronexus"]
    entities_to_test = ["neuronexus/allego_1/allego_2__uid0701-13-04-49.xdat.json"]


if __name__ == "__main__":
    unittest.main()



"""
Plexon2RawIO is a class to read Plexon PL2 files.

This IO is based on the pypl2lib module, see https://github.com/NeuralEnsemble/pypl2

On non-windows systems this IO requires Wine, see https://www.winehq.org/

Note on IO performance and memory consumption:
This IO is only partially lazy: To load a data chunk of an individual
signal channels, the complete data of that channel will be loaded into memory
and cached. This occurs as a limitation of the pypl2 package and might be improved
in future versions by exploiting the block-wise loading of signals.

To clear the currently cached signal data use the `clear_analogsignal_cache()` method.

Note on Neo header and spike channels:
The IO only considers enabled channels and will not list disabled channels in its header.

There is no 1-1 correspondence of PL2 spike channels and neo spike channels as each unit of
a PL2 spike channel will be represented as an individual neo spike channel.

Author: Julia Sprenger
"""

import pathlib
import warnings
import platform
import re

from urllib.request import urlopen
from datetime import datetime

import numpy as np


from ..baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _signal_buffer_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class Plexon2RawIO(BaseRawIO):
    """
    Class for "reading" data from a PL2 file

    Parameters
    ----------
    filename: str | Path
        The *.pl2 file to be loaded
    pl2_dll_file_path: str | Path | None, default: None
        The path to the necessary dll for loading pl2 files
        If None will find correct dll for architecture and if it does not exist will download it
    reading_attempts: int, default: 25
        Number of attempts to read the file before raising an error
        This opening process is somewhat unreliable and might fail occasionally. Adjust this higher
        if you encounter problems in opening the file.

    Notes
    -----
    * This IO is only partially lazy
    * The IO only considers enabled channels and will not list disabled channels in its header.
    * There is no 1-1 correspondence of PL2 spike channels and neo spike channels as each unit of
      a PL2 spike channel will be represented as an individual neo spike channel.

    Examples
    --------
    >>> import neo.rawio
    >>> r = neo.rawio.Plexon2RawIO(filename='my_data.pl2')
    >>> r.parse_header()
    >>> print(r)
    >>> raw_chunk = r.get_analogsignal_chunk(block_index=0,
                                             seg_index=0,
                                             i_start=0,
                                             i_stop=1024,
                                             stream_index=0,
                                             channel_indexes=range(10))
    >>> float_chunk = r.rescale_signal_raw_to_float(raw_chunk,
                                                    dtype='float64',
                                                    stream_index=0,
                                                    channel_indexes=[0, 3, 6])
    >>> spike_timestamp = r.get_spike_timestamps(spike_channel_index=0,
                                                 t_start=None,
                                                 t_stop=None)
    >>> spike_times = r.rescale_spike_timestamp(spike_timestamp, dtype='float64')
    >>> ev_timestamps, _, ev_labels = r.get_event_timestamps(event_channel_index=0)

    """

    extensions = ["pl2"]
    rawmode = "one-file"

    def __init__(self, filename, pl2_dll_file_path=None, reading_attempts=25):

        # signals, event and spiking data will be cached
        # cached signal data can be cleared using `clear_analogsignal_cache()()`
        self._analogsignal_cache = {}
        self._event_channel_cache = {}
        self._spike_channel_cache = {}

        # note that this filename is used in self._source_name
        self.filename = pathlib.Path(filename)

        if (not self.filename.exists()) or (not self.filename.is_file()):
            raise ValueError(f"{self.filename} is not a file.")

        BaseRawIO.__init__(self)

        # download default PL2 dll once if not yet available
        if pl2_dll_file_path is None:
            architecture = platform.architecture()[0]
            if architecture == "64bit" and platform.system() in ["Windows", "Darwin"]:
                file_name = "PL2FileReader64.dll"
            else:  # Apparently wine uses the 32 bit version in linux
                file_name = "PL2FileReader.dll"
            pl2_dll_folder = pathlib.Path.home() / ".plexon_dlls_for_neo"
            pl2_dll_folder.mkdir(exist_ok=True)
            pl2_dll_file_path = pl2_dll_folder / file_name

            if not pl2_dll_file_path.exists():
                url = f"https://raw.githubusercontent.com/Neuralensemble/pypl2/master/bin/{file_name}"
                dist = urlopen(url=url)

                with open(pl2_dll_file_path, "wb") as f:
                    warnings.warn(f"dll file does not exist, downloading plexon dll to {pl2_dll_file_path}")
                    f.write(dist.read())

        # Instantiate wrapper for Windows DLL
        from neo.rawio.plexon2rawio.pypl2.pypl2lib import PyPL2FileReader

        self.pl2reader = PyPL2FileReader(pl2_dll_file_path=pl2_dll_file_path)

        for attempt in range(reading_attempts):
            self.pl2reader.pl2_open_file(self.filename)

            # Verify the file handle is valid.
            if self.pl2reader._file_handle.value != 0:
                # File handle is valid, exit the loop early
                break
            else:
                if attempt == reading_attempts - 1:
                    self.pl2reader._print_error()
                    raise IOError(f"Opening {self.filename} failed after {reading_attempts} attempts.")

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        """
        Collect information about the file, construct neo header and provide annotations
        """

        # Scanning sources and populating signal channels at the same time. Sources have to have
        # same sampling rate and number of samples to belong to one stream.
        signal_channels = []
        channel_num_samples = []

        # We will build the stream ids based on the channel prefixes
        # The channel prefixes are the first characters of the channel names which have the following format:
        # WB{number}, FPX{number}, SPKCX{number}, AI{number}, etc
        # We will extract the prefix and use it as stream id
        regex_prefix_pattern = r"^\D+"  # Match any non-digit character at the beginning of the string

        for channel_index in range(self.pl2reader.pl2_file_info.m_TotalNumberOfAnalogChannels):
            achannel_info = self.pl2reader.pl2_get_analog_channel_info(channel_index)
            # only consider active channels
            if not (achannel_info.m_ChannelEnabled and achannel_info.m_ChannelRecordingEnabled):
                continue

            # assign to matching stream or create new stream based on signal characteristics
            rate = achannel_info.m_SamplesPerSecond
            num_samples = achannel_info.m_NumberOfValues
            channel_num_samples.append(num_samples)

            ch_name = achannel_info.m_Name.decode()
            chan_id = f"source{achannel_info.m_Source}.{achannel_info.m_Channel}"
            dtype = "int16"
            units = achannel_info.m_Units.decode()
            gain = achannel_info.m_CoeffToConvertToUnits
            offset = 0.0  # PL2 files don't contain information on signal offset

            channel_prefix = re.match(regex_prefix_pattern, ch_name).group(0)
            stream_id = channel_prefix
            buffer_id = ""
            signal_channels.append((ch_name, chan_id, rate, dtype, units, gain, offset, stream_id, buffer_id))

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)
        channel_num_samples = np.array(channel_num_samples)

        # We are using channel prefixes as stream_ids
        # The meaning of the channel prefixes was provided by a Plexon Engineer, see here:
        # https://github.com/NeuralEnsemble/python-neo/pull/1495#issuecomment-2184256894
        stream_id_to_stream_name = {
            "WB": "WB-Wideband",
            "FP": "FPl-Low Pass Filtered",
            "SP": "SPKC-High Pass Filtered",
            "AI": "AI-Auxiliary Input",
            "AIF": "AIF-Auxiliary Input Filtered",
        }

        unique_stream_ids = np.unique(signal_channels["stream_id"])
        signal_streams = []
        for stream_id in unique_stream_ids:
            # We are using the channel prefixes as ids
            # The users of plexon can modify the prefix of the channel names (e.g. `my_prefix` instead of `WB`).
            # In that case we use the channel prefix both as stream id and name
            stream_name = stream_id_to_stream_name.get(stream_id, stream_id)
            buffer_id = ""
            signal_streams.append((stream_name, stream_id, buffer_id))
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)
        # In plexon buffer is unknown
        signal_buffers = np.array([], dtype=_signal_buffer_dtype)

        self._stream_id_samples = {}
        self._stream_index_to_stream_id = {}
        for stream_index, stream_id in enumerate(signal_streams["id"]):
            # Keep a mapping from stream_index to stream_id
            self._stream_index_to_stream_id[stream_index] = stream_id

            # We extract the number of samples for each stream
            mask = signal_channels["stream_id"] == stream_id
            signal_num_samples = np.unique(channel_num_samples[mask])
            assert signal_num_samples.size == 1, "All channels in a stream must have the same number of samples"
            self._stream_id_samples[stream_id] = signal_num_samples[0]

        # pre-loading spike channel_data for later usage
        self._spike_channel_cache = {}
        spike_channels = []
        for c in range(self.pl2reader.pl2_file_info.m_TotalNumberOfSpikeChannels):
            schannel_info = self.pl2reader.pl2_get_spike_channel_info(c)

            # only consider active channels
            if not (schannel_info.m_ChannelEnabled and schannel_info.m_ChannelRecordingEnabled):
                continue

            # In a PL2 spike channel header, the field "m_NumberOfUnits" denotes the number
            # of units to which spikes detected on that channel have been assigned. It does
            # not account for unsorted spikes, i.e., spikes that have not been assigned to
            # a unit. It is Plexon's convention to assign unsorted spikes to channel_unit_id=0,
            # and sorted spikes to channel_unit_id's 1, 2, 3...etc. Therefore, for a given value of
            # m_NumberOfUnits, there are m_NumberOfUnits+1 channel_unit_ids to consider - 1
            # unsorted channel_unit_id (0) + the m_NumberOfUnits sorted channel_unit_ids.
            for channel_unit_id in range(schannel_info.m_NumberOfUnits + 1):
                unit_name = f"{schannel_info.m_Name.decode()}.{channel_unit_id}"
                unit_id = f"unit{schannel_info.m_Channel}.{channel_unit_id}"
                wf_units = schannel_info.m_Units
                wf_gain = schannel_info.m_CoeffToConvertToUnits
                wf_offset = 0.0  # A waveform offset is not provided in PL2 files
                wf_left_sweep = schannel_info.m_PreThresholdSamples
                wf_sampling_rate = schannel_info.m_SamplesPerSecond
                spike_channels.append(
                    (unit_name, unit_id, wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate)
                )

        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # creating event/epoch channel
        self._event_channel_cache = {}
        event_channels = []
        for i in range(self.pl2reader.pl2_file_info.m_NumberOfDigitalChannels):
            echannel_info = self.pl2reader.pl2_get_digital_channel_info(i)

            # only consider active channels
            if not (echannel_info.m_ChannelEnabled and echannel_info.m_ChannelRecordingEnabled):
                continue

            # event channels are characterized by (name, id, type), with type in ['event', 'epoch']
            channel_name = echannel_info.m_Name.decode()
            event_channels.append((channel_name, echannel_info.m_Channel, "event"))

        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # fill into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]  # It seems pl2 can only contain a single segment
        self.header["signal_buffers"] = signal_buffers
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()

        # this pprint lines really help for understand the nested (and complicated sometimes) dict
        # from pprint import pprint
        # pprint(self.raw_annotations)

        # Note: pl2_file_info.m_ReprocessorDateTime seems to be always empty.
        # To be checked against alternative pl2 reader.

        # Provide additional, recommended annotations for the final neo objects.
        block_index = 0
        bl_ann = self.raw_annotations["blocks"][block_index]
        bl_ann["name"] = f"Block containing PL2 data#{block_index}"
        bl_ann["file_origin"] = self.filename
        file_info = self.pl2reader.pl2_file_info
        block_info = {attr: getattr(file_info, attr) for attr, _ in file_info._fields_}

        # convert ctypes datetime objects to datetime.datetime objects for annotations
        from .pypl2.pypl2lib import tm

        for anno_key, anno_value in block_info.items():
            if isinstance(anno_value, tm):
                tmo = anno_value
                # invalid datetime information if year is <1
                if tmo.tm_year != 0:
                    microseconds = block_info["m_CreatorDateTimeMilliseconds"] * 1000
                    # tm_mon range is 0..11 https://cplusplus.com/reference/ctime/tm/
                    # python is 1..12 https://docs.python.org/3/library/datetime.html#datetime.datetime
                    # so month needs to be tm_mon+1; also tm_sec could cause problems in the case of leap
                    # seconds, but this is harder to defend against.
                    year = tmo.tm_year  # This has base 1900 in the c++ struct specification so we need to add 1900
                    year += 1900
                    dt = datetime(
                        year=year,
                        month=tmo.tm_mon + 1,
                        day=tmo.tm_mday,
                        hour=tmo.tm_hour,
                        minute=tmo.tm_min,
                        second=tmo.tm_sec,
                        microsecond=microseconds,
                    )
                    # ignoring daylight saving time information for now as timezone is unknown

                else:
                    dt = None

                block_info[anno_key] = dt

        bl_ann.update(block_info)
        for seg_index in range(1):
            seg_ann = bl_ann["segments"][seg_index]

            # some attributes don't apply to neo spike channels as these cover only a subpart of
            # the data of a PL2 spike channel
            spike_annotation_keys = [
                "m_SortEnabled",
                "m_SortRangeStart",
                "m_SortRangeEnd",
                "m_SourceTrodality",
                "m_OneBasedTrode",
                "m_OneBasedChannelInTrode",
                "m_Source",
                "m_Channel",
            ]
            for spike_channel_idx, spike_header in enumerate(self.header["spike_channels"]):
                schannel_name = spike_header["name"].split(".")[0]
                schannel_info = self.pl2reader.pl2_get_spike_channel_info_by_name(schannel_name)

                spiketrain_an = seg_ann["spikes"][spike_channel_idx]
                for key in spike_annotation_keys:
                    spiketrain_an[key] = getattr(schannel_info, key)

            event_annotation_keys = ["m_Source", "m_Channel", "m_Name"]
            for event_channel_idx, event_header in enumerate(self.header["event_channels"]):
                dchannel_name = event_header["name"]
                dchannel_info = self.pl2reader.pl2_get_digital_channel_info_by_name(dchannel_name)

                event_an = seg_ann["events"][event_channel_idx]
                for key in event_annotation_keys:
                    event_an[key] = getattr(dchannel_info, key)

            signal_array_annotation_keys = [
                "m_SourceTrodality",
                "m_OneBasedTrode",
                "m_OneBasedChannelInTrode",
                "m_Source",
                "m_Channel",
                "m_MaximumNumberOfFragments",
            ]

            for stream_idx, stream_header in enumerate(self.header["signal_streams"]):
                signal_array_annotations = {key: [] for key in signal_array_annotation_keys}
                stream_id = stream_header["id"]
                # extract values of individual signals

                stream_channel_mask = self.header["signal_channels"]["stream_id"] == stream_id
                for signal_idx, signal_header in enumerate(self.header["signal_channels"][stream_channel_mask]):
                    achannel_name = signal_header["name"]
                    achannel_info = self.pl2reader.pl2_get_analog_channel_info_by_name(achannel_name)
                    for key in signal_array_annotation_keys:
                        signal_array_annotations[key].append(getattr(achannel_info, key))

                seg_ann["signals"][stream_idx]["__array_annotations__"] = signal_array_annotations

    def _segment_t_start(self, block_index, seg_index):
        # this must return a float values in seconds
        return self.pl2reader.pl2_file_info.m_StartRecordingTime / self.pl2reader.pl2_file_info.m_TimestampFrequency

    def _segment_t_stop(self, block_index, seg_index):
        # this must return a float value in seconds
        end_time = (
            self.pl2reader.pl2_file_info.m_StartRecordingTime + self.pl2reader.pl2_file_info.m_DurationOfRecording
        )
        return float(end_time / self.pl2reader.pl2_file_info.m_TimestampFrequency)

    def _get_signal_size(self, block_index, seg_index, stream_index):
        stream_id = self._stream_index_to_stream_id[stream_index]
        num_samples = int(self._stream_id_samples[stream_id])
        return num_samples

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        # This returns the t_start of signals as a float value in seconds

        # TODO: Does the fragment_timestamp[0] need to be added here for digital signals?
        return self._segment_t_start(block_index, seg_index)

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        # this must return a signal chunk in a signal stream
        # limited with i_start/i_stop (can be None)
        # channel_indexes can be None (=all channel in the stream) or a list or numpy.array
        # This must return a numpy array 2D (even with one channel).
        # This must return the original dtype. No conversion here.
        # This must as fast as possible.
        # To speed up this call all preparatory calculations should be implemented
        # in _parse_header().

        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        stream_channels = self.header["signal_channels"][mask]

        n_channels = len(stream_channels)
        n_samples = self.get_signal_size(block_index, seg_index, stream_index)

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = n_samples

        if i_start < 0 or i_stop > n_samples:
            raise IndexError(f"Indexes ({i_start}, {i_stop}) out of range for signal with {n_samples} samples")

        # converting channel_indexes to array representation
        if channel_indexes is None:
            channel_indexes = np.arange(len(stream_channels), dtype="int")
        elif isinstance(channel_indexes, slice):
            channel_indexes = np.arange(len(stream_channels), dtype="int")[channel_indexes]
        else:
            channel_indexes = np.asarray(channel_indexes)

        # channel index sanity check
        if any(channel_indexes < 0) or any(channel_indexes >= n_channels):
            raise IndexError(f"Channel index out of range {channel_indexes} for stream with {n_channels} channels")

        nb_chan = len(channel_indexes)

        raw_signals = np.empty((i_stop - i_start, nb_chan), dtype="int16")
        for i, channel_idx in enumerate(channel_indexes):
            channel_name = stream_channels["name"][channel_idx]

            # use previously loaded channel data if possible
            if channel_name in self._analogsignal_cache:
                values = self._analogsignal_cache[channel_name]
            else:
                res = self.pl2reader.pl2_get_analog_channel_data_by_name(channel_name)
                fragment_timestamps, fragment_counts, values = res
                self._analogsignal_cache[channel_name] = values

            raw_signals[:, i] = values[i_start:i_stop]

        return raw_signals

    def clear_analogsignal_cache(self):
        for channel_name, values in self._analogsignal_cache.items():
            del values
        self._analogsignal_cache = {}

    def _spike_count(self, block_index, seg_index, spike_channel_index):
        channel_header = self.header["spike_channels"][spike_channel_index]
        channel_name, channel_unit_id = channel_header["name"].split(".")
        channel_unit_id = int(channel_unit_id)

        # loading spike channel data on demand when not already cached
        if channel_name not in self._spike_channel_cache:
            self._spike_channel_cache[channel_name] = self.pl2reader.pl2_get_spike_channel_data_by_name(channel_name)

        spike_timestamps, unit_ids, waveforms = self._spike_channel_cache[channel_name]
        nb_spikes = np.count_nonzero(unit_ids == channel_unit_id)

        return nb_spikes

    def _get_spike_timestamps(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        channel_header = self.header["spike_channels"][spike_channel_index]
        channel_name, channel_unit_id = channel_header["name"].split(".")
        channel_unit_id = int(channel_unit_id)

        # loading spike channel data on demand when not already cached
        if channel_name not in self._spike_channel_cache:
            self._spike_channel_cache[channel_name] = self.pl2reader.pl2_get_spike_channel_data_by_name(channel_name)

        spike_timestamps, unit_ids, waveforms = self._spike_channel_cache[channel_name]

        time_mask = self._get_timestamp_time_mask(t_start, t_stop, spike_timestamps)

        unit_mask = unit_ids[time_mask] == channel_unit_id
        spike_timestamps = spike_timestamps[time_mask][unit_mask]

        # spike timestamps are counted from the session start recording time
        spike_timestamps += self.pl2reader.pl2_file_info.m_StartRecordingTime

        return spike_timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        # must rescale to second a particular spike_timestamps
        # with a fixed dtype so the user can choose the precision they want
        spike_times = spike_timestamps.astype(dtype)
        spike_times /= self.pl2reader.pl2_file_info.m_TimestampFrequency
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        # this must return a 3D numpy array (nb_spike, nb_channel, nb_sample)
        # in the original dtype
        # this must be as fast as possible.
        # the same clip t_start/t_start must be used in _spike_timestamps()

        channel_header = self.header["spike_channels"][spike_channel_index]
        channel_name, channel_unit_id = channel_header["name"].split(".")

        # loading spike channel data on demand when not already cached
        if channel_name not in self._spike_channel_cache:
            self._spike_channel_cache[channel_name] = self.pl2reader.pl2_get_spike_channel_data_by_name(channel_name)

        spike_timestamps, unit_ids, waveforms = self._spike_channel_cache[channel_name]

        time_mask = self._get_timestamp_time_mask(t_start, t_stop, spike_timestamps)

        unit_mask = unit_ids[time_mask] == int(channel_unit_id)
        waveforms = waveforms[time_mask][unit_mask]

        # add tetrode dimension
        waveforms = np.expand_dims(waveforms, axis=1)
        return waveforms

    def _get_timestamp_time_mask(self, t_start, t_stop, timestamps):

        if t_start is not None or t_stop is not None:
            # restrict spikes to given limits (in seconds)
            timestamp_frequency = self.pl2reader.pl2_file_info.m_TimestampFrequency
            lim0 = int(t_start * timestamp_frequency)
            lim1 = int(t_stop * self.pl2reader.pl2_file_info.m_TimestampFrequency)

            # limits are with respect to segment t_start and not to time 0
            lim0 -= self.pl2reader.pl2_file_info.m_StartRecordingTime
            lim1 -= self.pl2reader.pl2_file_info.m_StartRecordingTime

            time_mask = (timestamps >= lim0) & (timestamps <= lim1)

        else:
            time_mask = slice(None, None)

        return time_mask

    def _event_count(self, block_index, seg_index, event_channel_index):

        channel_header = self.header["event_channels"][event_channel_index]
        channel_name = channel_header["name"]

        # loading event channel data on demand when not already cached
        if channel_name not in self._event_channel_cache:
            self._event_channel_cache[channel_name] = self.pl2reader.pl2_get_digital_channel_data_by_name(channel_name)

        event_times, values = self._event_channel_cache[channel_name]

        return len(event_times)

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        # the main difference between spike channel and event channel
        # is that for here we have 3 numpy array timestamp, durations, labels
        # durations must be None for 'event'
        # label must a dtype ='U'

        channel_header = self.header["event_channels"][event_channel_index]
        channel_name = channel_header["name"]

        # loading event channel data on demand when not already cached
        if channel_name not in self._event_channel_cache:
            self._event_channel_cache[channel_name] = self.pl2reader.pl2_get_digital_channel_data_by_name(channel_name)

        event_times, labels = self._event_channel_cache[channel_name]
        labels = np.asarray(labels, dtype="U")

        time_mask = self._get_timestamp_time_mask(t_start, t_stop, event_times)

        # events don't have a duration. Epochs are not supported
        durations = None

        # event timestamps are counted from the session start recording time
        return_times = event_times[time_mask] + self.pl2reader.pl2_file_info.m_StartRecordingTime

        return return_times, durations, labels[time_mask]

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        # must rescale to second a particular event_timestamps
        # with a fixed dtype so the user can choose the precision he want.

        event_times = event_timestamps.astype(dtype)
        event_times /= self.pl2reader.pl2_file_info.m_TimestampFrequency
        return event_times

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        durations = raw_duration.astype(dtype)
        durations /= self.pl2reader.pl2_file_info.m_TimestampFrequency
        return durations

    def close(self):
        self.pl2reader.pl2_close_all_files()



import unittest

from neo.rawio.openephysbinaryrawio import OpenEphysBinaryRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO

import numpy as np


class TestOpenEphysBinaryRawIO(BaseTestRawIO, unittest.TestCase):
    rawioclass = OpenEphysBinaryRawIO
    entities_to_download = ["openephysbinary"]
    entities_to_test = [
        "openephysbinary/v0.5.3_two_neuropixels_stream",
        "openephysbinary/v0.4.4.1_with_video_tracking",
        "openephysbinary/v0.5.x_two_nodes",
        "openephysbinary/v0.6.x_neuropixels_multiexp_multistream",
        "openephysbinary/v0.6.x_neuropixels_with_sync",
        "openephysbinary/v0.6.x_neuropixels_missing_folders",
        "openephysbinary/neural_and_non_neural_data_mixed",
    ]

    def test_sync(self):
        rawio_with_sync = OpenEphysBinaryRawIO(
            self.get_local_path("openephysbinary/v0.6.x_neuropixels_with_sync"), load_sync_channel=True
        )
        rawio_with_sync.parse_header()
        stream_name = [s_name for s_name in rawio_with_sync.header["signal_streams"]["name"] if "AP" in s_name][0]
        stream_index = list(rawio_with_sync.header["signal_streams"]["name"]).index(stream_name)

        # AP stream has 385 channels
        chunk = rawio_with_sync.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 385

        rawio_no_sync = OpenEphysBinaryRawIO(
            self.get_local_path("openephysbinary/v0.6.x_neuropixels_with_sync"), load_sync_channel=False
        )
        rawio_no_sync.parse_header()

        # AP stream has 384 channels
        chunk = rawio_no_sync.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 384

    def test_sync_channel_access(self):
        """Test that sync channels can be accessed as separate streams when load_sync_channel=False."""
        rawio = OpenEphysBinaryRawIO(
            self.get_local_path("openephysbinary/v0.6.x_neuropixels_with_sync"), load_sync_channel=False
        )
        rawio.parse_header()

        # Find sync channel streams
        sync_stream_names = [s_name for s_name in rawio.header["signal_streams"]["name"] if "SYNC" in s_name]
        assert len(sync_stream_names) > 0, "No sync channel streams found"

        # Get the stream index for the first sync channel
        sync_stream_index = list(rawio.header["signal_streams"]["name"]).index(sync_stream_names[0])

        # Check that we can access the sync channel data
        chunk = rawio.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=sync_stream_index
        )

        # Sync channel should have only one channel
        assert chunk.shape[1] == 1, f"Expected sync channel to have 1 channel, got {chunk.shape[1]}"

    def test_no_sync(self):
        # requesting sync channel when there is none raises an error
        with self.assertRaises(ValueError):
            rawio_no_sync = OpenEphysBinaryRawIO(
                self.get_local_path("openephysbinary/v0.6.x_neuropixels_multiexp_multistream"), load_sync_channel=True
            )
            rawio_no_sync.parse_header()

    def test_missing_folders(self):
        # missing folders should raise an error
        with self.assertWarns(UserWarning):
            rawio = OpenEphysBinaryRawIO(
                self.get_local_path("openephysbinary/v0.6.x_neuropixels_missing_folders"), load_sync_channel=False
            )
            rawio.parse_header()

    def test_multiple_ttl_events_parsing(self):
        rawio = OpenEphysBinaryRawIO(
            self.get_local_path("openephysbinary/v0.6.x_neuropixels_with_sync"), load_sync_channel=False
        )
        rawio.parse_header()
        rawio.header = rawio.header
        # Testing co
        # This is the TTL events from the NI Board channel
        ttl_events = rawio._evt_streams[0][0][1]
        assert "rising" in ttl_events.keys()
        assert "labels" in ttl_events.keys()
        assert "durations" in ttl_events.keys()
        assert "timestamps" in ttl_events.keys()

        # Check that durations of different event streams are correctly parsed:
        assert np.allclose(ttl_events["durations"][ttl_events["labels"] == "1"], 0.5, atol=0.001)
        assert np.allclose(ttl_events["durations"][ttl_events["labels"] == "6"], 0.025, atol=0.001)
        assert np.allclose(ttl_events["durations"][ttl_events["labels"] == "7"], 0.016666, atol=0.001)

    def test_separating_stream_for_non_neural_data(self):
        rawio = OpenEphysBinaryRawIO(
            self.get_local_path("openephysbinary/neural_and_non_neural_data_mixed"), load_sync_channel=False
        )
        rawio.parse_header()
        # Check that the non-neural data stream is correctly separated
        assert len(rawio.header["signal_streams"]["name"]) == 2
        assert rawio.header["signal_streams"]["name"].tolist() == ["Rhythm_FPGA-100.0", "Rhythm_FPGA-100.0_ADC"]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.rawio.spikeglxrawio
"""

import unittest

from neo.rawio.spikeglxrawio import SpikeGLXRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO
import numpy as np


class TestSpikeGLXRawIO(BaseTestRawIO, unittest.TestCase):
    rawioclass = SpikeGLXRawIO
    entities_to_download = ["spikeglx"]
    entities_to_test = [
        "spikeglx/Noise4Sam_g0",
        "spikeglx/TEST_20210920_0_g0",
        # this is only g0 multi index
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI0/5-19-2022-CI0_g0",
        # this is only g1 multi index
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI0/5-19-2022-CI0_g1",
        # this mix both multi gate and multi trigger (and also multi probe)
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI0",
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI1",
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI2",
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI3",
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI4",
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI5",
        # different sync/sybset options with commercial NP2
        "spikeglx/NP2_with_sync",
        "spikeglx/NP2_no_sync",
        "spikeglx/NP2_subset_with_sync",
        # NP-ultra
        "spikeglx/np_ultra_stub",
        # Filename changed by the user, multi-dock
        "spikeglx/multi_probe_multi_dock_multi_shank_filename_without_info",
        # CatGT
        "spikeglx/multi_trigger_multi_gate/CatGT/CatGT-A",
        "spikeglx/multi_trigger_multi_gate/CatGT/CatGT-B",
        "spikeglx/multi_trigger_multi_gate/CatGT/CatGT-C",
        "spikeglx/multi_trigger_multi_gate/CatGT/CatGT-D",
        "spikeglx/multi_trigger_multi_gate/CatGT/CatGT-E",
        "spikeglx/multi_trigger_multi_gate/CatGT/Supercat-A",
    ]

    def test_loading_only_one_probe_in_multi_probe_scenario(self):
        from pathlib import Path

        local_path_multi_probe_path = Path(
            self.get_local_path("spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI0")
        )
        gate_folder_path = local_path_multi_probe_path / "5-19-2022-CI0_g0"
        probe_folder_path = gate_folder_path / "5-19-2022-CI0_g0_imec1"

        rawio = SpikeGLXRawIO(probe_folder_path)
        rawio.parse_header()

        expected_stream_names = ["imec1.ap", "imec1.lf"]
        actual_stream_names = rawio.header["signal_streams"]["name"].tolist()
        assert (
            actual_stream_names == expected_stream_names
        ), f"Expected {expected_stream_names}, but got {actual_stream_names}"

    def test_with_location(self):
        rawio = SpikeGLXRawIO(self.get_local_path("spikeglx/Noise4Sam_g0"), load_channel_location=True)
        rawio.parse_header()
        # one of the stream have channel location
        have_location = []
        for sig_anotations in rawio.raw_annotations["blocks"][0]["segments"][0]["signals"]:
            have_location.append("channel_location_0" in sig_anotations["__array_annotations__"])
        assert any(have_location)

    def test_sync(self):
        rawio_with_sync = SpikeGLXRawIO(self.get_local_path("spikeglx/NP2_with_sync"), load_sync_channel=True)
        rawio_with_sync.parse_header()
        stream_index = list(rawio_with_sync.header["signal_streams"]["name"]).index("imec0.ap")

        # AP stream has 385 channels
        chunk = rawio_with_sync.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 385

        rawio_no_sync = SpikeGLXRawIO(self.get_local_path("spikeglx/NP2_with_sync"), load_sync_channel=False)
        rawio_no_sync.parse_header()

        # AP stream has 384 channels
        chunk = rawio_no_sync.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 384

    def test_no_sync(self):
        # requesting sync channel when there is none raises an error
        with self.assertRaises(ValueError):
            rawio_no_sync = SpikeGLXRawIO(self.get_local_path("spikeglx/NP2_no_sync"), load_sync_channel=True)
            rawio_no_sync.parse_header()

    def test_subset_with_sync(self):
        rawio_sub = SpikeGLXRawIO(self.get_local_path("spikeglx/NP2_subset_with_sync"), load_sync_channel=True)
        rawio_sub.parse_header()
        stream_index = list(rawio_sub.header["signal_streams"]["name"]).index("imec0.ap")

        # AP stream has 121 channels
        chunk = rawio_sub.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 121

        rawio_sub_no_sync = SpikeGLXRawIO(self.get_local_path("spikeglx/NP2_subset_with_sync"), load_sync_channel=False)
        rawio_sub_no_sync.parse_header()
        # AP stream has 120 channels
        chunk = rawio_sub_no_sync.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 120

    def test_nidq_digital_channel(self):
        rawio_digital = SpikeGLXRawIO(self.get_local_path("spikeglx/DigitalChannelTest_g0"))
        rawio_digital.parse_header()
        # This data should have 8 event channels
        assert np.shape(rawio_digital.header["event_channels"])[0] == 8

        # Channel 0 in this data will have sync pulses at 1 Hz, let's confirm that
        all_events = rawio_digital.get_event_timestamps(0, 0, 0)
        on_events = np.where(all_events[2] == "XD0 ON")
        on_ts = all_events[0][on_events]
        on_ts_scaled = rawio_digital.rescale_event_timestamp(on_ts)
        on_diff = np.diff(on_ts_scaled)
        atol = 0.001
        assert np.allclose(on_diff, 1, atol=atol)

    def test_t_start_reading(self):
        """Test that t_start values are correctly read for all streams and segments."""

        # Expected t_start values for each stream and segment
        expected_t_starts = {
            "imec0.ap": {0: 15.319535472007237, 1: 15.339535431281986, 2: 21.284723325294053, 3: 21.3047232845688},
            "imec1.ap": {0: 15.319554693264516, 1: 15.339521518106308, 2: 21.284735282142822, 3: 21.304702106984614},
            "imec0.lf": {0: 15.3191688060872, 1: 15.339168765361949, 2: 21.284356659374016, 3: 21.304356618648765},
            "imec1.lf": {0: 15.319321358082725, 1: 15.339321516521915, 2: 21.284568614155827, 3: 21.30456877259502},
        }

        # Initialize the RawIO
        rawio = SpikeGLXRawIO(self.get_local_path("spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI4"))
        rawio.parse_header()

        # Get list of stream names
        stream_names = rawio.header["signal_streams"]["name"]

        # Test t_start for each stream and segment
        for stream_name, expected_values in expected_t_starts.items():
            # Get stream index
            stream_index = list(stream_names).index(stream_name)

            # Check each segment
            for seg_index, expected_t_start in expected_values.items():
                actual_t_start = rawio.get_signal_t_start(block_index=0, seg_index=seg_index, stream_index=stream_index)

                # Use numpy.testing for proper float comparison
                np.testing.assert_allclose(
                    actual_t_start,
                    expected_t_start,
                    rtol=1e-9,
                    atol=1e-9,
                    err_msg=f"Mismatch in t_start for stream '{stream_name}', segment {seg_index}",
                )


if __name__ == "__main__":
    unittest.main()



"""
Common tools that are useful for neo.io object tests
"""

import logging
import os
import importlib.util


logger = logging.getLogger("neo.test")


def can_use_network():
    """
    Return True if network access is allowed
    """

    # env variable for local dev
    if os.environ.get("NEO_TESTS_NO_NETWORK", False):
        return False

    # check for datalad presence
    datalad_spec = importlib.util.find_spec("datalad")
    if datalad_spec is not None:
        HAVE_DATALAD = True
    else:
        HAVE_DATALAD = False

    if not HAVE_DATALAD:
        return False

    return True



import unittest
import numpy as np
from pathlib import Path

from neo.rawio.intanrawio import IntanRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestIntanRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = IntanRawIO
    entities_to_download = ["intan"]
    entities_to_test = [
        "intan/intan_rhs_test_1.rhs",  # Format header-attached
        "intan/intan_rhd_test_1.rhd",  # Format header-attached
        "intan/rhs_fpc_multistim_240514_082243/rhs_fpc_multistim_240514_082243.rhs",  # Format header-attached newer version
        "intan/intan_fpc_test_231117_052630/info.rhd",  # Format one-file-per-channel
        "intan/intan_fps_test_231117_052500/info.rhd",  # Format one file per signal
        "intan/intan_fpc_rhs_test_240329_091637/info.rhs",  # Format one-file-per-channel
        "intan/intan_fps_rhs_test_240329_091536/info.rhs",  # Format one-file-per-signal
        "intan/rhd_fpc_multistim_240514_082044/info.rhd",  # Multiple digital channels one-file-per-channel rhd
        "intan/rhs_stim_data_single_file_format/intanTestFile.rhs",  # header-attached rhs data with stimulus current
        "intan/test_fcs_dc_250327_154333/info.rhs",  # this is an example of only having dc amp rather than amp files
        "intan/test_fpc_stim_250327_151617/info.rhs",  # wrong files names Heberto will fix naimgin in the future
    ]

    def test_annotations(self):

        intan_reader = IntanRawIO(filename=self.get_local_path("intan/intan_rhd_test_1.rhd"))
        intan_reader.parse_header()

        raw_annotations = intan_reader.raw_annotations
        annotations = raw_annotations["blocks"][0]["segments"][0]  # Intan is mono segment
        signal_annotations = annotations["signals"][0]  # As in the other exmaples, annotaions are duplicated

        # Scalar annotations
        exepcted_annotations = {
            "intan_version": "1.5",
            "desired_impedance_test_frequency": 1000.0,
            "desired_upper_bandwidth": 7500.0,
            "note1": "",
            "notch_filter_mode": 1,
            "notch_filter": False,
            "nb_signal_group": 7,
            "dsp_enabled": 1,
            "actual_impedance_test_frequency": 1000.0,
            "desired_lower_bandwidth": 0.1,
            "note3": "",
            "actual_dsp_cutoff_frequency": 1.165828,
            "desired_dsp_cutoff_frequency": 1.0,
            "actual_lower_bandwidth": 0.0945291,
            "eval_board_mode": 0,
            "note2": "",
            "num_temp_sensor_channels": 0,
        }

        for key in exepcted_annotations:
            if isinstance(exepcted_annotations[key], float):
                self.assertAlmostEqual(signal_annotations[key], exepcted_annotations[key], places=2)
            else:
                self.assertEqual(signal_annotations[key], exepcted_annotations[key])

        # Array annotations
        signal_array_annotations = signal_annotations["__array_annotations__"]
        np.testing.assert_array_equal(signal_array_annotations["native_order"][:10], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        np.testing.assert_array_equal(
            signal_array_annotations["spike_scope_digital_edge_polarity"][:10], [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
        )
        np.testing.assert_array_equal(signal_array_annotations["board_stream_num"][:10], [0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def test_correct_reading_one_file_per_channel_amplifiers(self):
        "Issue: https://github.com/NeuralEnsemble/python-neo/issues/1599"
        # Test reading of one-file-per-channel format file. The channels should match the raw files
        file_path = Path(self.get_local_path("intan/intan_fpc_test_231117_052630/info.rhd"))
        intan_reader = IntanRawIO(filename=file_path)
        intan_reader.parse_header()

        # This should be the folder where the files of all the channels are stored
        folder_path = file_path.parent

        # The paths are named as amp-A-000.dat, amp-A-001.dat, amp-A-002.dat, ...
        amplifier_file_paths = [path for path in folder_path.iterdir() if "amp" in path.name]
        channel_names = [path.name[4:-4] for path in amplifier_file_paths]

        amplifier_stream_index = 0
        for channel_name, amplifier_file_path in zip(channel_names, amplifier_file_paths):
            data_raw = np.fromfile(amplifier_file_path, dtype=np.int16).squeeze()
            data_from_neo = intan_reader.get_analogsignal_chunk(
                channel_ids=[channel_name], stream_index=amplifier_stream_index
            ).squeeze()
            np.testing.assert_allclose(data_raw, data_from_neo)

    def test_correct_reading_one_file_per_channel_rhs_stim(self):
        "Zach request for testing that the channel order is correct"
        # Test reading of one-file-per-channel format file. The channels should match the raw files
        file_path = Path(self.get_local_path("intan/test_fpc_stim_250327_151617/info.rhs"))
        intan_reader = IntanRawIO(filename=file_path)
        intan_reader.parse_header()

        # This should be the folder where the files of all the channels are stored
        folder_path = file_path.parent

        # The paths for the stim channels are stim-A-000.dat, stim-A-001.dat, stim-A-002.dat,
        # Whereas the ids are A-001_STIM, A-002_STIM, A-003_STIM, etc
        stim_file_paths = [path for path in folder_path.iterdir() if "stim" in path.name]
        channel_ids = [f"{p.stem[5:]}_STIM" for p in stim_file_paths]

        stim_stream_index = 2
        for channel_id, amplifier_file_path in zip(channel_ids, stim_file_paths):
            data_raw = np.fromfile(amplifier_file_path, dtype=np.uint16)
            decoded_data = intan_reader._decode_current_from_stim_data(data_raw, 0, data_raw.shape[0])
            data_from_neo = intan_reader.get_analogsignal_chunk(
                channel_ids=[channel_id], stream_index=stim_stream_index
            ).squeeze()
            np.testing.assert_allclose(decoded_data, data_from_neo)

    def test_correct_decoding_of_stimulus_current(self):
        # See https://github.com/NeuralEnsemble/python-neo/pull/1660 for discussion
        # See https://gin.g-node.org/NeuralEnsemble/ephy_testing_data/src/master/intan/README.md#rhs_stim_data_single_file_format
        # For a description of the data

        file_path = Path(self.get_local_path("intan/rhs_stim_data_single_file_format/intanTestFile.rhs"))
        intan_reader = IntanRawIO(filename=file_path)
        intan_reader.parse_header()

        signal_streams = intan_reader.header["signal_streams"]
        stream_ids = signal_streams["id"].tolist()
        stream_index = stream_ids.index("11")
        sampling_rate = intan_reader.get_signal_sampling_rate(stream_index=stream_index)
        sig_chunk = intan_reader.get_analogsignal_chunk(stream_index=stream_index, channel_ids=["D-016_STIM"])
        final_stim = intan_reader.rescale_signal_raw_to_float(
            sig_chunk, stream_index=stream_index, channel_ids=["D-016_STIM"]
        )

        # This contains only the first pulse and I got this by visual inspection
        data_to_test = final_stim[200:250]

        positive_pulse_size = np.max(data_to_test).item()
        negative_pulse_size = np.min(data_to_test).item()

        expected_value = 60 * 1e-6  # 60 microamperes

        # Assert is close float
        assert np.isclose(positive_pulse_size, expected_value)
        assert np.isclose(negative_pulse_size, -expected_value)

        # Check that negative pulse is leading
        argmin = np.argmin(data_to_test)
        argmax = np.argmax(data_to_test)
        assert argmin < argmax

        # Check that the negative pulse is 200 us long
        negative_pulse_frames = np.where(data_to_test > 0)[0]
        number_of_negative_frames = negative_pulse_frames.size
        duration_of_negative_pulse = number_of_negative_frames / sampling_rate

        expected_duration = 200 * 1e-6  # 400 microseconds / 2
        assert np.isclose(duration_of_negative_pulse, expected_duration)

        # Check that the positive pulse is 200 us long
        positive_pulse_frames = np.where(data_to_test > 0)[0]
        number_of_positive_frames = positive_pulse_frames.size
        duration_of_positive_pulse = number_of_positive_frames / sampling_rate
        expected_duration = 200 * 1e-6  # 400 microseconds / 2

        assert np.isclose(duration_of_positive_pulse, expected_duration)

    def test_correct_decoding_of_stimulus_current(self):
        # See https://github.com/NeuralEnsemble/python-neo/pull/1660 for discussion
        # See https://gin.g-node.org/NeuralEnsemble/ephy_testing_data/src/master/intan/README.md#rhs_stim_data_single_file_format
        # For a description of the data

        file_path = Path(self.get_local_path("intan/rhs_stim_data_single_file_format/intanTestFile.rhs"))
        intan_reader = IntanRawIO(filename=file_path)
        intan_reader.parse_header()

        signal_streams = intan_reader.header["signal_streams"]
        stream_ids = signal_streams["id"].tolist()
        stream_index = stream_ids.index("11")
        sampling_rate = intan_reader.get_signal_sampling_rate(stream_index=stream_index)
        sig_chunk = intan_reader.get_analogsignal_chunk(stream_index=stream_index, channel_ids=["D-016_STIM"])
        final_stim = intan_reader.rescale_signal_raw_to_float(
            sig_chunk, stream_index=stream_index, channel_ids=["D-016_STIM"]
        )

        # This contains only the first pulse and I got this by visual inspection
        data_to_test = final_stim[200:250]

        positive_pulse_size = np.max(data_to_test).item()
        negative_pulse_size = np.min(data_to_test).item()

        expected_value = 60 * 1e-6  # 60 microamperes

        # Assert is close float
        assert np.isclose(positive_pulse_size, expected_value)
        assert np.isclose(negative_pulse_size, -expected_value)

        # Check that negative pulse is leading
        argmin = np.argmin(data_to_test)
        argmax = np.argmax(data_to_test)
        assert argmin < argmax

        # Check that the negative pulse is 200 us long
        negative_pulse_frames = np.where(data_to_test > 0)[0]
        number_of_negative_frames = negative_pulse_frames.size
        duration_of_negative_pulse = number_of_negative_frames / sampling_rate

        expected_duration = 200 * 1e-6  # 400 microseconds / 2
        assert np.isclose(duration_of_negative_pulse, expected_duration)

        # Check that the positive pulse is 200 us long
        positive_pulse_frames = np.where(data_to_test > 0)[0]
        number_of_positive_frames = positive_pulse_frames.size
        duration_of_positive_pulse = number_of_positive_frames / sampling_rate
        expected_duration = 200 * 1e-6  # 400 microseconds / 2

        assert np.isclose(duration_of_positive_pulse, expected_duration)


if __name__ == "__main__":
    unittest.main()



"""
Class for reading the old data format from Plexon
acquisition system (.plx)

Note that Plexon now use a new format PL2 which is NOT
supported by this IO.

Compatible with versions 100 to 106.
Other versions have not been tested.

This IO is developed thanks to the header file downloadable from:
http://www.plexon.com/software-downloads

This IO was rewritten in 2017 and this was a huge pain because
the underlying file format is really inefficient.
The rewrite is now based on numpy dtype and not on Python struct.
This should be faster.
If one day, somebody use it, consider to offer me a beer.


Author: Samuel Garcia

"""

import datetime
from collections import OrderedDict
import re

import numpy as np

try:
    from tqdm import tqdm, trange

    HAVE_TQDM = True
except:
    HAVE_TQDM = False

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _signal_buffer_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)

from neo.core.baseneo import NeoReadWriteError


class PlexonRawIO(BaseRawIO):
    extensions = ["plx"]
    rawmode = "one-file"

    def __init__(self, filename="", progress_bar=True):
        """

        Class for reading non-pl2 plexon files

        Parameters
        ----------
        filename: str, default: ''
            The *.plx file to be loaded
        progress_bar: bool, default True
            Display progress bar using tqdm (if installed) when parsing the file.

        Notes
        -----
        * Compatible with versions 100 to 106. Other versions have not been tested.
        * Note that Plexon now use a new format PL2 which is NOT supported by this IO.

        Examples
        --------
        >>> import neo.rawio
        >>> r = neo.rawio.PlexonRawIO(filename='data.plx')
        >>> r.parse_header()
        >>> print(r)

        """
        BaseRawIO.__init__(self)
        self.filename = filename
        self.progress_bar = HAVE_TQDM and progress_bar

    def _source_name(self):
        return self.filename

    def _parse_header(self):

        # global header
        with open(self.filename, "rb") as fid:
            global_header = read_as_dict(fid, GlobalHeader)

        rec_datetime = datetime.datetime(
            global_header["Year"],
            global_header["Month"],
            global_header["Day"],
            global_header["Hour"],
            global_header["Minute"],
            global_header["Second"],
        )

        # dsp channels header = spikes and waveforms
        nb_unit_chan = int(global_header["NumDSPChannels"])
        offset1 = np.dtype(GlobalHeader).itemsize
        dspChannelHeaders = np.memmap(
            self.filename, dtype=DspChannelHeader, mode="r", offset=offset1, shape=(nb_unit_chan,)
        )

        # event channel header
        nb_event_chan = int(global_header["NumEventChannels"])
        offset2 = offset1 + np.dtype(DspChannelHeader).itemsize * nb_unit_chan
        eventHeaders = np.memmap(
            self.filename,
            dtype=EventChannelHeader,
            mode="r",
            offset=offset2,
            shape=(nb_event_chan,),
        )

        # slow channel header = signal
        nb_sig_chan = int(global_header["NumSlowChannels"])
        offset3 = offset2 + np.dtype(EventChannelHeader).itemsize * nb_event_chan
        slowChannelHeaders = np.memmap(
            self.filename, dtype=SlowChannelHeader, mode="r", offset=offset3, shape=(nb_sig_chan,)
        )

        offset4 = offset3 + np.dtype(SlowChannelHeader).itemsize * nb_sig_chan

        # locate data blocks and group them by type and channel
        block_pos = {
            1: {c: [] for c in dspChannelHeaders["Channel"]},
            4: {c: [] for c in eventHeaders["Channel"]},
            5: {c: [] for c in slowChannelHeaders["Channel"]},
        }
        data = self._memmap = np.memmap(self.filename, dtype="u1", offset=0, mode="r")
        pos = offset4

        # Create a tqdm object with a total of len(data) and an initial value of 0 for offset
        if self.progress_bar:
            progress_bar = tqdm(total=len(data), initial=0, desc="Parsing data blocks", leave=True)

        while pos < data.size:
            bl_header = data[pos : pos + 16].view(DataBlockHeader)[0]
            number_of_waveforms = int(bl_header["NumberOfWaveforms"])
            number_of_words_in_waveform = int(bl_header["NumberOfWordsInWaveform"])
            length = (number_of_waveforms * number_of_words_in_waveform * 2) + 16
            bl_type = int(bl_header["Type"])
            chan_id = int(bl_header["Channel"])
            block_pos[bl_type][chan_id].append(pos)
            pos += length

            # Update tqdm with the number of bytes processed in this iteration
            if self.progress_bar:
                progress_bar.update(length)  # This was clever, Sam : )

        if self.progress_bar:
            progress_bar.close()

        upper_byte_of_5_byte_timestamp = int(bl_header["UpperByteOf5ByteTimestamp"])
        bl_header_timestamp = int(bl_header["TimeStamp"])
        self._last_timestamps = upper_byte_of_5_byte_timestamp * 2**32 + bl_header_timestamp

        # ... and finalize them in self._data_blocks
        # for a faster access depending on type (1, 4, 5)
        self._data_blocks = {}
        dt_base = [("pos", "int64"), ("timestamp", "int64"), ("size", "int64")]
        dtype_by_bltype = {
            # Spikes and waveforms
            1: np.dtype(
                dt_base
                + [
                    ("unit_id", "uint16"),
                    ("n1", "uint16"),
                    ("n2", "uint16"),
                ]
            ),
            # Events
            4: np.dtype(
                dt_base
                + [
                    ("label", "uint16"),
                ]
            ),
            # Signals
            5: np.dtype(
                dt_base
                + [
                    ("cumsum", "int64"),
                ]
            ),
        }
        if self.progress_bar:
            bl_loop = tqdm(block_pos, desc="Finalizing data blocks", leave=True)
        else:
            bl_loop = block_pos
        for bl_type in bl_loop:
            self._data_blocks[bl_type] = {}
            if self.progress_bar:
                chan_loop = tqdm(
                    block_pos[bl_type],
                    desc=f"Finalizing data blocks for type {bl_type}",
                    leave=True,
                )
            else:
                chan_loop = block_pos[bl_type]
            for chan_id in chan_loop:
                positions = block_pos[bl_type][chan_id]
                dt = dtype_by_bltype[bl_type]
                data_block = np.empty((len(positions)), dtype=dt)
                for index, pos in enumerate(positions):
                    bl_header = data[pos : pos + 16].view(DataBlockHeader)[0]

                    # To avoid overflow errors when doing arithmetic operations on numpy scalars
                    np_scalar_to_python_scalar = lambda x: x.item() if isinstance(x, np.generic) else x
                    bl_header = {key: np_scalar_to_python_scalar(bl_header[key]) for key in bl_header.dtype.names}

                    current_upper_byte_of_5_byte_timestamp = int(bl_header["UpperByteOf5ByteTimestamp"])
                    current_bl_timestamp = int(bl_header["TimeStamp"])
                    timestamp = current_upper_byte_of_5_byte_timestamp * 2**32 + current_bl_timestamp
                    n1 = bl_header["NumberOfWaveforms"]
                    n2 = bl_header["NumberOfWordsInWaveform"]
                    sample_count = n1 * n2

                    data_block["pos"][index] = pos + 16
                    data_block["timestamp"][index] = timestamp
                    data_block["size"][index] = sample_count * 2

                    if bl_type == 1:  # Spikes and waveforms
                        data_block["unit_id"][index] = bl_header["Unit"]
                        data_block["n1"][index] = n1
                        data_block["n2"][index] = n2
                    elif bl_type == 4:  # Events
                        data_block["label"][index] = bl_header["Unit"]
                    elif bl_type == 5:  # Signals
                        if data_block.size > 0:
                            # cumulative sum of sample index for fast access to chunks
                            if index == 0:
                                data_block["cumsum"][index] = 0
                            else:
                                data_block["cumsum"][index] = data_block["cumsum"][index - 1] + sample_count

                self._data_blocks[bl_type][chan_id] = data_block

        # signals channels
        source_id = []

        # Scanning sources and populating signal channels at the same time. Sources have to have
        # same sampling rate and number of samples to belong to one stream.
        signal_channels = []
        channel_num_samples = []

        # We will build the stream ids based on the channel prefixes
        # The channel prefixes are the first characters of the channel names which have the following format:
        # WB{number}, FPX{number}, SPKCX{number}, AI{number}, etc
        # We will extract the prefix and use it as stream id
        regex_prefix_pattern = r"^\D+"  # Match any non-digit character at the beginning of the string

        if self.progress_bar:
            chan_loop = trange(nb_sig_chan, desc="Parsing signal channels", leave=True)
        else:
            chan_loop = range(nb_sig_chan)
        for chan_index in chan_loop:
            slow_channel_headers = slowChannelHeaders[chan_index]

            # To avoid overflow errors when doing arithmetic operations on numpy scalars
            np_scalar_to_python_scalar = lambda x: x.item() if isinstance(x, np.generic) else x
            slow_channel_headers = {
                key: np_scalar_to_python_scalar(slow_channel_headers[key]) for key in slow_channel_headers.dtype.names
            }

            name = slow_channel_headers["Name"].decode("utf8")
            chan_id = slow_channel_headers["Channel"]
            length = self._data_blocks[5][chan_id]["size"].sum() // 2
            if length == 0:
                continue  # channel not added
            source_id.append(slow_channel_headers["SrcId"])
            channel_num_samples.append(length)
            sampling_rate = float(slow_channel_headers["ADFreq"])
            sig_dtype = "int16"
            units = ""  # I don't know units
            if global_header["Version"] in [100, 101]:
                gain = 5000.0 / (2048 * slow_channel_headers["Gain"] * 1000.0)
            elif global_header["Version"] in [102]:
                gain = 5000.0 / (2048 * slow_channel_headers["Gain"] * slow_channel_headers["PreampGain"])
            elif global_header["Version"] >= 103:
                gain = global_header["SlowMaxMagnitudeMV"] / (
                    0.5
                    * (2 ** global_header["BitsPerSpikeSample"])
                    * slow_channel_headers["Gain"]
                    * slow_channel_headers["PreampGain"]
                )
            offset = 0.0

            channel_prefix = re.match(regex_prefix_pattern, name).group(0)
            stream_id = channel_prefix
            buffer_id = ""
            signal_channels.append(
                (name, str(chan_id), sampling_rate, sig_dtype, units, gain, offset, stream_id, buffer_id)
            )

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        # no buffer here because block are splitted by channel
        signal_buffers = np.array([], dtype=_signal_buffer_dtype)

        if signal_channels.size == 0:
            signal_streams = np.array([], dtype=_signal_stream_dtype)

        else:
            # Detect streams
            channel_num_samples = np.asarray(channel_num_samples)
            # We are using channel prefixes as stream_ids
            # The meaning of the channel prefixes was provided by a Plexon Engineer, see here:
            # https://github.com/NeuralEnsemble/python-neo/pull/1495#issuecomment-2184256894
            stream_id_to_stream_name = {
                "WB": "WB-Wideband",
                "FP": "FPl-Low Pass Filtered",
                "SP": "SPKC-High Pass Filtered",
                "AI": "AI-Auxiliary Input",
                "AIF": "AIF-Auxiliary Input Filtered",
            }

            unique_stream_ids = np.unique(signal_channels["stream_id"])
            signal_streams = []
            for stream_id in unique_stream_ids:
                # We are using the channel prefixes as ids
                # The users of plexon can modify the prefix of the channel names (e.g. `my_prefix` instead of `WB`).
                # In that case we use the channel prefix both as stream id and name
                buffer_id = ""
                stream_name = stream_id_to_stream_name.get(stream_id, stream_id)
                buffer_id = ""
                signal_streams.append((stream_name, stream_id, buffer_id))

            signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

            self._stream_id_samples = {}
            self._stream_id_sampling_frequency = {}
            self._stream_index_to_stream_id = {}
            for stream_index, stream_id in enumerate(signal_streams["id"]):
                # Keep a mapping from stream_index to stream_id
                self._stream_index_to_stream_id[stream_index] = stream_id

                mask = signal_channels["stream_id"] == stream_id

                signal_num_samples = np.unique(channel_num_samples[mask])
                if signal_num_samples.size > 1:
                    raise NeoReadWriteError(f"Channels in stream {stream_id} don't have the same number of samples")
                self._stream_id_samples[stream_id] = signal_num_samples[0]

                signal_sampling_frequency = np.unique(signal_channels[mask]["sampling_rate"])
                if signal_sampling_frequency.size > 1:
                    raise NeoReadWriteError(f"Channels in stream {stream_id} don't have the same sampling frequency")
                self._stream_id_sampling_frequency[stream_id] = signal_sampling_frequency[0]

        self._global_ssampling_rate = global_header["ADFrequency"]

        # Determine number of units per channels
        self.internal_unit_ids = []
        for chan_id, data_clock in self._data_blocks[1].items():
            unit_ids = np.unique(data_clock["unit_id"])
            for unit_id in unit_ids:
                self.internal_unit_ids.append((chan_id, unit_id))

        # Spikes channels
        spike_channels = []
        if self.progress_bar:
            unit_loop = tqdm(
                enumerate(self.internal_unit_ids),
                desc="Parsing spike channels",
                leave=True,
            )
        else:
            unit_loop = enumerate(self.internal_unit_ids)

        for unit_index, (chan_id, unit_id) in unit_loop:
            channel_index = np.nonzero(dspChannelHeaders["Channel"] == chan_id)[0][0]
            dsp_channel_headers = dspChannelHeaders[channel_index]

            name = dsp_channel_headers["Name"].decode("utf8")
            _id = f"ch{chan_id}#{unit_id}"
            wf_units = ""
            if global_header["Version"] < 103:
                wf_gain = 3000.0 / (2048 * dsp_channel_headers["Gain"] * 1000.0)
            elif 103 <= global_header["Version"] < 105:
                wf_gain = global_header["SpikeMaxMagnitudeMV"] / (
                    0.5 * 2.0 ** (global_header["BitsPerSpikeSample"]) * dsp_channel_headers["Gain"] * 1000.0
                )
            elif global_header["Version"] >= 105:
                wf_gain = global_header["SpikeMaxMagnitudeMV"] / (
                    0.5
                    * 2.0 ** (global_header["BitsPerSpikeSample"])
                    * dsp_channel_headers["Gain"]
                    * global_header["SpikePreAmpGain"]
                )
            wf_offset = 0.0
            wf_left_sweep = -1  # DONT KNOWN
            wf_sampling_rate = global_header["WaveformFreq"]
            spike_channels.append((name, _id, wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate))
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # Event channels
        event_channels = []
        if self.progress_bar:
            chan_loop = trange(nb_event_chan, desc="Parsing event channels", leave=True)
        else:
            chan_loop = range(nb_event_chan)
        for chan_index in chan_loop:
            h = eventHeaders[chan_index]
            chan_id = h["Channel"]
            name = h["Name"].decode("utf8")
            event_channels.append((name, chan_id, "event"))
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # fill into header dict
        self.header = {
            "nb_block": 1,
            "nb_segment": [1],
            "signal_buffers": signal_buffers,
            "signal_streams": signal_streams,
            "signal_channels": signal_channels,
            "spike_channels": spike_channels,
            "event_channels": event_channels,
        }

        # Annotations
        self._generate_minimal_annotations()
        bl_annotations = self.raw_annotations["blocks"][0]
        seg_annotations = bl_annotations["segments"][0]
        for d in (bl_annotations, seg_annotations):
            d["rec_datetime"] = rec_datetime
            d["plexon_version"] = global_header["Version"]

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = float(self._last_timestamps) / self._global_ssampling_rate
        if hasattr(self, "__stream_id_samples"):
            for stream_id in self._stream_id_samples.keys():
                t_stop_sig = self._stream_id_samples[stream_id] / self._stream_id_sampling_frequency[stream_id]
                t_stop = max(t_stop, t_stop_sig)
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        stream_id = self._stream_index_to_stream_id[stream_index]
        return self._stream_id_samples[stream_id]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        signal_channels = self.header["signal_channels"]
        signal_streams = self.header["signal_streams"]
        stream_id = signal_streams[stream_index]["id"]

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._stream_id_samples[stream_id]

        mask = signal_channels["stream_id"] == stream_id
        signal_channels = signal_channels[mask]
        if channel_indexes is not None:
            signal_channels = signal_channels[channel_indexes]
        channel_ids = signal_channels["id"]

        raw_signals = np.zeros((i_stop - i_start, channel_ids.size), dtype="int16")
        for c, channel_id in enumerate(channel_ids):
            chan_id = np.int32(channel_id)

            data_blocks = self._data_blocks[5][chan_id]

            # loop over data blocks and get chunks
            bl0 = np.searchsorted(data_blocks["cumsum"], i_start, side="right") - 1
            bl1 = np.searchsorted(data_blocks["cumsum"], i_stop, side="right")
            ind = 0
            for bl in range(bl0, bl1):
                ind0 = data_blocks[bl]["pos"]
                ind1 = data_blocks[bl]["size"] + ind0
                data = self._memmap[ind0:ind1].view("int16")
                if bl == bl1 - 1:
                    # right border
                    # be careful that bl could be both bl0 and bl1!!
                    border = data.size - (i_stop - data_blocks[bl]["cumsum"])
                    data = data[:-border]
                if bl == bl0:
                    # left border
                    border = i_start - data_blocks[bl]["cumsum"]
                    data = data[border:]
                raw_signals[ind : data.size + ind, c] = data
                ind += data.size

        return raw_signals

    def _get_internal_mask(self, data_block, t_start, t_stop):
        timestamps = data_block["timestamp"]

        if t_start is None:
            lim0 = 0
        else:
            lim0 = int(t_start * self._global_ssampling_rate)

        if t_stop is None:
            lim1 = self._last_timestamps
        else:
            lim1 = int(t_stop * self._global_ssampling_rate)

        keep = (timestamps >= lim0) & (timestamps <= lim1)

        return keep

    def _spike_count(self, block_index, seg_index, unit_index):
        chan_id, unit_id = self.internal_unit_ids[unit_index]
        data_block = self._data_blocks[1][chan_id]
        nb_spike = np.sum(data_block["unit_id"] == unit_id)
        return nb_spike

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        chan_id, unit_id = self.internal_unit_ids[unit_index]
        data_block = self._data_blocks[1][chan_id]

        keep = self._get_internal_mask(data_block, t_start, t_stop)
        keep &= data_block["unit_id"] == unit_id
        spike_timestamps = data_block[keep]["timestamp"]

        return spike_timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        spike_times = spike_timestamps.astype(dtype)
        spike_times /= self._global_ssampling_rate
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        chan_id, unit_id = self.internal_unit_ids[unit_index]
        data_block = self._data_blocks[1][chan_id]

        n1 = data_block["n1"][0]
        n2 = data_block["n2"][0]

        keep = self._get_internal_mask(data_block, t_start, t_stop)
        keep &= data_block["unit_id"] == unit_id

        data_block = data_block[keep]
        nb_spike = data_block.size

        waveforms = np.zeros((nb_spike, n1, n2), dtype="int16")
        for i, db in enumerate(data_block):
            ind0 = db["pos"]
            ind1 = db["size"] + ind0
            data = self._memmap[ind0:ind1].view("int16").reshape(n1, n2)
            waveforms[i, :, :] = data

        return waveforms

    def _event_count(self, block_index, seg_index, event_channel_index):
        chan_id = int(self.header["event_channels"][event_channel_index]["id"])
        nb_event = self._data_blocks[4][chan_id].size
        return nb_event

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        chan_id = int(self.header["event_channels"][event_channel_index]["id"])
        data_block = self._data_blocks[4][chan_id]
        keep = self._get_internal_mask(data_block, t_start, t_stop)

        db = data_block[keep]
        timestamps = db["timestamp"]
        labels = db["label"].astype("U")
        durations = None

        return timestamps, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        event_times = event_timestamps.astype(dtype)
        event_times /= self._global_ssampling_rate
        return event_times


def read_as_dict(fid, dtype, offset=None):
    """
    Given a file descriptor
    and a numpy.dtype of the binary struct return a dict.
    Make conversion for strings.
    """
    if offset is not None:
        fid.seek(offset)
    dt = np.dtype(dtype)
    h = np.frombuffer(fid.read(dt.itemsize), dt)[0]
    info = OrderedDict()
    for k in dt.names:
        v = h[k]

        if dt[k].kind == "S":
            v = v.decode("utf8")
            v = v.replace("\x03", "")
            v = v.replace("\x00", "")

        info[k] = v.item() if isinstance(v, np.generic) else v
    return info


GlobalHeader = [
    ("MagicNumber", "uint32"),
    ("Version", "int32"),
    ("Comment", "S128"),
    ("ADFrequency", "int32"),
    ("NumDSPChannels", "int32"),
    ("NumEventChannels", "int32"),
    ("NumSlowChannels", "int32"),
    ("NumPointsWave", "int32"),
    ("NumPointsPreThr", "int32"),
    ("Year", "int32"),
    ("Month", "int32"),
    ("Day", "int32"),
    ("Hour", "int32"),
    ("Minute", "int32"),
    ("Second", "int32"),
    ("FastRead", "int32"),
    ("WaveformFreq", "int32"),
    ("LastTimestamp", "float64"),
    # version >103
    ("Trodalness", "uint8"),
    ("DataTrodalness", "uint8"),
    ("BitsPerSpikeSample", "uint8"),
    ("BitsPerSlowSample", "uint8"),
    ("SpikeMaxMagnitudeMV", "uint16"),
    ("SlowMaxMagnitudeMV", "uint16"),
    # version 105
    ("SpikePreAmpGain", "uint16"),
    # version 106
    ("AcquiringSoftware", "S18"),
    ("ProcessingSoftware", "S18"),
    ("Padding", "S10"),
    # all version
    ("TSCounts", "int32", (650,)),
    ("WFCounts", "int32", (650,)),
    ("EVCounts", "int32", (512,)),
]

DspChannelHeader = [
    ("Name", "S32"),
    ("SIGName", "S32"),
    ("Channel", "int32"),
    ("WFRate", "int32"),
    ("SIG", "int32"),
    ("Ref", "int32"),
    ("Gain", "int32"),
    ("Filter", "int32"),
    ("Threshold", "int32"),
    ("Method", "int32"),
    ("NUnits", "int32"),
    ("Template", "uint16", (320,)),
    ("Fit", "int32", (5,)),
    ("SortWidth", "int32"),
    ("Boxes", "uint16", (40,)),
    ("SortBeg", "int32"),
    # version 105
    ("Comment", "S128"),
    # version 106
    ("SrcId", "uint8"),
    ("reserved", "uint8"),
    ("ChanId", "uint16"),
    ("Padding", "int32", (10,)),
]

EventChannelHeader = [
    ("Name", "S32"),
    ("Channel", "int32"),
    # version 105
    ("Comment", "S128"),
    # version 106
    ("SrcId", "uint8"),
    ("reserved", "uint8"),
    ("ChanId", "uint16"),
    ("Padding", "int32", (32,)),
]

SlowChannelHeader = [
    ("Name", "S32"),
    ("Channel", "int32"),
    ("ADFreq", "int32"),
    ("Gain", "int32"),
    ("Enabled", "int32"),
    ("PreampGain", "int32"),
    # version 104
    ("SpikeChannel", "int32"),
    # version 105
    ("Comment", "S128"),
    # version 106
    ("SrcId", "uint8"),
    ("reserved", "uint8"),
    ("ChanId", "uint16"),
    ("Padding", "int32", (27,)),
]

DataBlockHeader = [
    ("Type", "uint16"),
    ("UpperByteOf5ByteTimestamp", "uint16"),
    ("TimeStamp", "int32"),
    ("Channel", "uint16"),
    ("Unit", "uint16"),
    ("NumberOfWaveforms", "uint16"),
    ("NumberOfWordsInWaveform", "uint16"),
]  # 16 bytes



import unittest
import numpy as np

from neo.rawio.neuronexusrawio import NeuroNexusRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO


class TestNeuroNexusRawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = NeuroNexusRawIO
    entities_to_download = ["neuronexus"]
    entities_to_test = ["neuronexus/allego_1/allego_2__uid0701-13-04-49.xdat.json"]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.rawio.mearecrawio

"""

import unittest

from neo.rawio.plexon2rawio import Plexon2RawIO

from neo.test.rawiotest.common_rawio_test import BaseTestRawIO

from numpy.testing import assert_equal

try:
    from neo.rawio.plexon2rawio.pypl2 import pypl2lib

    HAVE_PYPL2 = True
except (ImportError, TimeoutError):
    HAVE_PYPL2 = False


@unittest.skipUnless(HAVE_PYPL2, "requires pypl package and all its dependencies")
class TestPlexon2RawIO(
    BaseTestRawIO,
    unittest.TestCase,
):
    rawioclass = Plexon2RawIO
    entities_to_download = ["plexon"]
    entities_to_test = ["plexon/4chDemoPL2.pl2", "plexon/NC16FPSPKEVT_1m.pl2"]

    def test_check_enabled_flags(self):
        """
        This test loads a 1-minute PL2 file with 16 channels' each
        of field potential (FP), and spike (SPK) data. The channels
        cycle through 4 possible combinations of m_ChannelEnabled
        and m_ChannelRecordingEnabled - (True, True), (True, False),
        (False, True), and (False, False). With 16 channels for each
        source, each combination of flags occurs 4 times. Only the
        first combination (True, True) causes data to be recorded to
        disk. Therefore, we expect the following channels to be loaded by
        Neo: FP01, FP05, FP09, FP13, SPK01, SPK05, SPK09, and SPK13.

        Note: the file contains event (EVT) data as well. Although event
        channel headers do contain m_ChannelEnabled and m_ChannelRecording-
        Enabled flags, the UI for recording PL2 files does not expose any
        controls by which these flags can be changed from (True, True).
        Therefore, no test for event channels is necessary here.
        """

        # Load data from NC16FPSPKEVT_1m.pl2, a 1-minute PL2 recording containing
        # 16-channels' each of field potential (FP), spike (SPK), and event (EVT)
        # data.
        reader = Plexon2RawIO(filename=self.get_local_path("plexon/NC16FPSPKEVT_1m.pl2"))
        reader.parse_header()

        # Check that the names of the loaded signal channels match what we expect
        signal_channel_names = reader.header["signal_channels"]["name"].tolist()
        expected_signal_channel_names = ["FP01", "FP05", "FP09", "FP13"]
        assert_equal(signal_channel_names, expected_signal_channel_names)

        # Check that the names of the loaded spike channels match what we expect
        spike_channel_names = reader.header["spike_channels"]["name"].tolist()
        expected_spike_channel_names = ["SPK01.0", "SPK05.0", "SPK09.0", "SPK13.0"]
        assert_equal(spike_channel_names, expected_spike_channel_names)


if __name__ == "__main__":
    unittest.main()



"""
Support for intan tech rhd and rhs files.

These 2 formats are more or less the same but:
  * some variance in headers.
  * rhs amplifier is more complex because the optional DC channel

RHS supported version 1.0
RHD supported version  1.0 1.1 1.2 1.3 2.0 3.0, 3.1
RHD headerless binary support 3.x
RHS headerless binary support 3.x


See:
  * http://intantech.com/files/Intan_RHD2000_data_file_formats.pdf
  * http://intantech.com/files/Intan_RHS2000_data_file_formats.pdf


Author: Samuel Garcia (Initial), Zach McKenzie & Heberto Mayorquin (Updates)

"""

from pathlib import Path
from packaging.version import Version
import warnings

import numpy as np

from neo.core import NeoReadWriteError

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _signal_buffer_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class IntanRawIO(BaseRawIO):
    """
    Class for reading rhd and rhs Intan data

    Parameters
    ----------
    filename: str, default: ''
         name of the 'rhd' or 'rhs' data/header file
    ignore_integrity_checks: bool, default: False
        If True, data that violates integrity assumptions will be loaded. At the moment the only integrity
        check we perform is that timestamps are continuous. Setting this to True will ignore this check and set
        the attribute `discontinuous_timestamps` to True if the timestamps are not continous. This attribute can be checked
        after parsing the header to see if the timestamps are continuous or not.

    Notes
    -----
    * The Intan reader can handle two file formats 'rhd' and 'rhs'. It will automatically
      check for the file extension and will gather the header information based on the
      extension. Additionally it functions with RHS v 1.0 and v 3.x and RHD 1.0, 1.1, 1.2, 1.3, 2.0,
      3.x files.

    * The Intan reader can also handle the headerless binary formats 'one-file-per-signal' and
      'one-file-per-channel' which have a header file called 'info.rhd' or 'info.rhs' and a series
      of binary files with the '.dat' suffix

    * Intan files contain amplifier channels labeled 'A', 'B' 'C' or 'D' for the 512 recorder
      or 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H' for the 1024 recorder system
      depending on the port in which they were recorded along (stored in stream_id '0') with the following
      additional streams.

    0: 'RHD2000 amplifier channel'
    1: 'RHD2000 auxiliary input channel',
    2: 'RHD2000 supply voltage channel',
    3: 'USB board ADC input channel',
    4: 'USB board digital input channel',
    5: 'USB board digital output channel'

    And for RHS:

    0: 'RHS2000 amplfier channel'
    3: 'USB board ADC input channel',
    4: 'USB board ADC output channel',
    5: 'USB board digital input channel',
    6: 'USB board digital output channel',
    10: 'DC Amplifier channel',
    11: 'Stim channel',

    * We currently implement digital data demultiplexing so that if digital streams are requested they are
      returned as arrays of 1s and 0s.

    * We also do stim data decoding which returns the stim data as an int16 of appropriate magnitude. Please
      use `rescale_signal_raw_to_float` to obtain stim data in amperes.


    Examples
    --------
    >>> import neo.rawio
    >>> # for a header-attached file
    >>> reader = neo.rawio.IntanRawIO(filename='data.rhd')
    >>> # for the other formats we point to the info.rhd
    >>> reader = neo.rawioIntanRawIO(filename='info.rhd')
    >>> reader.parse_header()
    >>> raw_chunk = reader.get_analogsignal_chunk(block_index=0,
                                                  seg_index=0
                                                  stream_index=0)
    >>> float_chunk = reader.rescale_signal_raw_to_float(raw_chunk, stream_index=0)

    """

    extensions = ["rhd", "rhs", "dat"]
    rawmode = "one-file"

    def __init__(self, filename="", ignore_integrity_checks=False):

        BaseRawIO.__init__(self)
        self.filename = Path(filename)
        self.ignore_integrity_checks = ignore_integrity_checks
        self.discontinuous_timestamps = False

    def _source_name(self):
        return self.filename

    def _parse_header(self):

        self.filename = Path(self.filename)

        # Input checks
        if not self.filename.is_file():
            raise FileNotFoundError(f"{self.filename} does not exist")

        if not self.filename.suffix in [".rhd", ".rhs"]:
            raise ValueError(f"{self.filename} is not a valid Intan file. Expected .rhd or .rhs extension")

        # see comment below for RHD which explains the division between file types
        if self.filename.suffix == ".rhs":
            if self.filename.name == "info.rhs":
                if any((self.filename.parent / file).exists() for file in stream_name_to_file_name_rhs.values()):
                    self.file_format = "one-file-per-signal"
                    raw_file_paths_dict = create_one_file_per_signal_dict_rhs(dirname=self.filename.parent)
                else:
                    self.file_format = "one-file-per-channel"
                    raw_file_paths_dict = create_one_file_per_channel_dict_rhs(dirname=self.filename.parent)
            else:
                self.file_format = "header-attached"

            (
                self._global_info,
                self._ordered_channel_info,
                memmap_data_dtype,
                header_size,
                self._block_size,
                channel_number_dict,
            ) = read_rhs(self.filename, self.file_format)

        # 3 possibilities for rhd files, one combines the header and the data in the same file with suffix `rhd` while
        # the other two separates the data from the header which is always called `info.rhd`
        # attached to the actual binary file with data
        elif self.filename.suffix == ".rhd":
            if self.filename.name == "info.rhd":
                # first we have one-file-per-signal which is where one neo stream/file is saved as .dat files
                if any((self.filename.parent / file).exists() for file in stream_name_to_file_name_rhd.values()):
                    self.file_format = "one-file-per-signal"
                    raw_file_paths_dict = create_one_file_per_signal_dict_rhd(dirname=self.filename.parent)
                # then there is one-file-per-channel where each channel in a neo stream is in its own .dat file
                else:
                    self.file_format = "one-file-per-channel"
                    raw_file_paths_dict = create_one_file_per_channel_dict_rhd(dirname=self.filename.parent)
            # finally the format with the header-attached to the binary file as one giant file
            else:
                self.file_format = "header-attached"

            (
                self._global_info,
                self._ordered_channel_info,
                memmap_data_dtype,
                header_size,
                self._block_size,
                channel_number_dict,
            ) = read_rhd(self.filename, self.file_format)

        #############################################
        # memmap the raw data for each format type
        #############################################
        # if header-attached there is one giant memory-map
        if self.file_format == "header-attached":
            self._raw_data = np.memmap(self.filename, dtype=memmap_data_dtype, mode="r", offset=header_size)

        # for 'one-file-per-signal' we have one memory map / neo stream
        # based on https://github.com/NeuralEnsemble/python-neo/issues/1556 bug in versions 0.13.1, .2, .3
        elif self.file_format == "one-file-per-signal":
            self._raw_data = {}
            for stream_name, stream_dtype in memmap_data_dtype.items():
                num_channels = channel_number_dict[stream_name]
                file_path = raw_file_paths_dict[stream_name]
                size_in_bytes = file_path.stat().st_size
                dtype_size = np.dtype(stream_dtype).itemsize
                n_samples = size_in_bytes // (dtype_size * num_channels)
                signal_stream_memmap = np.memmap(
                    file_path,
                    dtype=stream_dtype,
                    mode="r",
                    shape=(n_samples, num_channels),
                    order="C",
                )
                self._raw_data[stream_name] = signal_stream_memmap

        # for one-file-per-channel we have one memory map / channel stored as a list / neo stream
        elif self.file_format == "one-file-per-channel":
            self._raw_data = {}
            for stream_name, stream_dtype in memmap_data_dtype.items():
                channel_file_paths = raw_file_paths_dict[stream_name]
                channel_memmap_list = [np.memmap(fp, dtype=stream_dtype, mode="r") for fp in channel_file_paths]
                self._raw_data[stream_name] = channel_memmap_list

        # Data Integrity checks
        # strictness of check is controlled by ignore_integrity_checks
        # which is set at __init__
        self._assert_timestamp_continuity()

        # signals
        signal_channels = []
        # This is used to calculate bit masks
        self.native_channel_order = {}
        for chan_info in self._ordered_channel_info:
            name = chan_info["custom_channel_name"]
            channel_id = chan_info["native_channel_name"]
            sig_dtype = chan_info["dtype"]
            stream_id = str(chan_info["signal_type"])
            # buffer will be handled later bepending the format
            buffer_id = ""
            signal_channels.append(
                (
                    name,
                    channel_id,
                    chan_info["sampling_rate"],
                    sig_dtype,
                    chan_info["units"],
                    chan_info["gain"],
                    chan_info["offset"],
                    stream_id,
                    buffer_id,
                )
            )
            self.native_channel_order[channel_id] = chan_info["native_order"]
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        stream_ids = np.unique(signal_channels["stream_id"])
        signal_streams = np.zeros(stream_ids.size, dtype=_signal_stream_dtype)

        buffer_ids = []
        # we need to sort the data because the string of stream_index 10 is mis-sorted.
        signal_streams["id"] = sorted(stream_ids, key=lambda x: int(x))
        for stream_index, stream_id in enumerate(signal_streams["id"]):
            if self.filename.suffix == ".rhd":
                name = stream_id_to_stream_name_rhd.get(stream_id, "")
            else:
                name = stream_id_to_stream_name_rhs.get(stream_id, "")
            signal_streams["name"][stream_index] = name
            # zach I need you help here
            if self.file_format == "header-attached":
                buffer_id = ""
            elif self.file_format == "one-file-per-signal":
                buffer_id = stream_id
                buffer_ids.append(buffer_id)
            elif self.file_format == "one-file-per-channel":
                buffer_id = ""

            signal_streams["buffer_id"][stream_index] = buffer_id

            # set buffer_id to channels
            if buffer_id != "":
                mask = signal_channels["stream_id"] == stream_id
                signal_channels["buffer_id"][mask] = buffer_id

        # depending the format we can have buffer_id or not
        signal_buffers = np.zeros(len(buffer_ids), dtype=_signal_buffer_dtype)
        signal_buffers["id"] = buffer_ids

        self._max_sampling_rate = np.max(signal_channels["sampling_rate"])

        # if header is attached we need to incorporate our block size to get signal length
        if self.file_format == "header-attached":
            self._max_sigs_length = self._raw_data.size * self._block_size
        # for one-file-per-signal we just take the size which will give n_samples for each
        # signal stream and then we just take the longest one
        elif self.file_format == "one-file-per-signal":
            self._max_sigs_length = max([raw_data.size for raw_data in self._raw_data.values()])
        # for one-file-per-channel we do the same as for one-file-per-signal, but since they
        # are in a list we just take the first channel in each list of channels
        else:
            self._max_sigs_length = max([raw_data[0].size for raw_data in self._raw_data.values()])

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # fill into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_buffers"] = signal_buffers
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # Extract annotations from the format
        self._generate_minimal_annotations()
        bl_annotations = self.raw_annotations["blocks"][0]
        seg_annotations = bl_annotations["segments"][0]

        for signal_annotation in seg_annotations["signals"]:
            # Add global annotations
            signal_annotation["intan_version"] = (
                f"{self._global_info['major_version']}." f"{self._global_info['minor_version']}"
            )
            global_keys_to_skip = [
                "major_version",
                "minor_version",
                "sampling_rate",
                "magic_number",
                "reference_channel",
            ]
            global_keys_to_annotate = set(self._global_info.keys()) - set(global_keys_to_skip)
            for key in global_keys_to_annotate:
                signal_annotation[key] = self._global_info[key]

            reference_channel = self._global_info.get("reference_channel", None)
            # Following the pdf specification
            reference_channel = "hardware" if reference_channel == "n/a" else reference_channel

            # Add channel annotations
            array_annotations = signal_annotation["__array_annotations__"]
            channel_ids = array_annotations["channel_ids"]

            # TODO refactor ordered channel dict to make this easier
            # Use this to find which elements of the ordered channels correspond to the current signal
            signal_type = int(signal_annotation["stream_id"])
            channel_info = next((info for info in self._ordered_channel_info if info["signal_type"] == signal_type))
            channel_keys_to_skip = [
                "signal_type",
                "custom_channel_name",
                "native_channel_name",
                "gain",
                "offset",
                "channel_enabled",
                "dtype",
                "units",
                "sampling_rate",
            ]

            channel_keys_to_annotate = set(channel_info.keys()) - set(channel_keys_to_skip)
            properties_dict = {key: [] for key in channel_keys_to_annotate}
            for channel_id in channel_ids:
                matching_info = next(
                    info for info in self._ordered_channel_info if info["native_channel_name"] == channel_id
                )
                for key in channel_keys_to_annotate:
                    properties_dict[key].append(matching_info[key])

            for key in channel_keys_to_annotate:
                array_annotations[key] = properties_dict[key]

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._max_sigs_length / self._max_sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):

        stream_name = self.header["signal_streams"][stream_index]["name"][:]

        if self.file_format == "header-attached":
            stream_id = self.header["signal_streams"][stream_index]["id"]
            mask = self.header["signal_channels"]["stream_id"] == stream_id
            signal_channels = self.header["signal_channels"][mask]
            channel_ids = signal_channels["id"]

            stream_is_digital = stream_name in digital_stream_names
            field_name = stream_name if stream_is_digital else channel_ids[0]

            size = self._raw_data[field_name].size

        # one-file-per-signal is (n_samples, n_channels)
        elif self.file_format == "one-file-per-signal":
            size = self._raw_data[stream_name].shape[0]
        # one-file-per-channel is (n_samples) so pull from list
        else:
            size = self._raw_data[stream_name][0].size

        return size

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._get_signal_size(block_index, seg_index, stream_index)

        if channel_indexes is None:
            channel_indexes = slice(None)

        if self.file_format == "header-attached":
            sigs_chunk = self._get_analogsignal_chunk_header_attached(
                i_start,
                i_stop,
                stream_index,
                channel_indexes,
            )
        elif self.file_format == "one-file-per-signal":
            sigs_chunk = self._get_analogsignal_chunk_one_file_per_signal(
                i_start,
                i_stop,
                stream_index,
                channel_indexes,
            )
        else:
            sigs_chunk = self._get_analogsignal_chunk_one_file_per_channel(
                i_start,
                i_stop,
                stream_index,
                channel_indexes,
            )

        return sigs_chunk

    def _get_analogsignal_chunk_header_attached(self, i_start, i_stop, stream_index, channel_indexes):

        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        signal_channels = self.header["signal_channels"][mask]
        channel_ids = signal_channels["id"][channel_indexes]

        stream_name = self.header["signal_streams"][stream_index]["name"][:]
        stream_is_digital = stream_name in digital_stream_names
        stream_is_stim = stream_name == "Stim channel"

        field_name = stream_name if stream_is_digital else channel_ids[0]

        shape = self._raw_data[field_name].shape
        dtype = self._raw_data[field_name].dtype

        # This is False for Temperature and timestamps
        multiple_samples_per_block = len(shape) == 2

        # In the header attached case the data for each channel comes interleaved in blocks
        # To avoid unecessary memory access we can calculate the blocks we need to access beforehand:
        if multiple_samples_per_block:
            block_size = shape[1]
            block_start = i_start // block_size
            block_stop = i_stop // block_size + 1
            sl0 = i_start % block_size
            sl1 = sl0 + (i_stop - i_start)

        if not stream_is_digital:
            # For all streams raw_data is a structured memmap with a field for each channel_id
            sigs_chunk = np.zeros((i_stop - i_start, len(channel_ids)), dtype=dtype)
            for chunk_index, channel_id in enumerate(channel_ids):
                data_chan = self._raw_data[channel_id]

                if multiple_samples_per_block:
                    sigs_chunk[:, chunk_index] = data_chan[block_start:block_stop].flatten()[sl0:sl1]
                else:
                    sigs_chunk[:, chunk_index] = data_chan[i_start:i_stop]

            if stream_is_stim:
                sigs_chunk = self._decode_current_from_stim_data(sigs_chunk, 0, sigs_chunk.shape[0])

        else:
            # For digital data the channels come interleaved in a single field so we need to demultiplex
            digital_raw_data = self._raw_data[field_name].flatten()
            sigs_chunk = self._demultiplex_digital_data(digital_raw_data, channel_ids, i_start, i_stop)
        return sigs_chunk

    def _get_analogsignal_chunk_one_file_per_channel(self, i_start, i_stop, stream_index, channel_indexes):

        stream_name = self.header["signal_streams"][stream_index]["name"][:]
        signal_data_memmap_list = self._raw_data[stream_name]
        stream_is_stim = stream_name == "Stim channel"

        channel_indexes_are_slice = isinstance(channel_indexes, slice)
        if channel_indexes_are_slice:
            num_channels = len(signal_data_memmap_list)
            start = channel_indexes.start or 0
            stop = channel_indexes.stop or num_channels
            step = channel_indexes.step or 1
            channel_indexes = range(start, stop, step)

        # We get the dtype from the first channel
        first_channel_index = channel_indexes[0]
        dtype = signal_data_memmap_list[first_channel_index].dtype
        sigs_chunk = np.zeros((i_stop - i_start, len(channel_indexes)), dtype=dtype)

        for chunk_index, channel_index in enumerate(channel_indexes):
            channel_memmap = signal_data_memmap_list[channel_index]
            sigs_chunk[:, chunk_index] = channel_memmap[i_start:i_stop]

        # If this is stim data, we need to decode the current values
        if stream_is_stim:
            sigs_chunk = self._decode_current_from_stim_data(sigs_chunk, 0, sigs_chunk.shape[0])

        return sigs_chunk

    def _get_analogsignal_chunk_one_file_per_signal(self, i_start, i_stop, stream_index, channel_indexes):

        stream_name = self.header["signal_streams"][stream_index]["name"][:]
        raw_data = self._raw_data[stream_name]

        stream_is_digital = stream_name in digital_stream_names
        stream_is_stim = stream_name == "Stim channel"

        if stream_is_digital:
            stream_id = self.header["signal_streams"][stream_index]["id"]
            mask = self.header["signal_channels"]["stream_id"] == stream_id
            signal_channels = self.header["signal_channels"][mask]
            channel_ids = signal_channels["id"][channel_indexes]

            output = self._demultiplex_digital_data(raw_data, channel_ids, i_start, i_stop)
        elif stream_is_stim:
            output = self._decode_current_from_stim_data(raw_data, i_start, i_stop)
            output = output[:, channel_indexes]
        else:
            output = raw_data[i_start:i_stop, channel_indexes]

        return output

    def _demultiplex_digital_data(self, raw_digital_data, channel_ids, i_start, i_stop):
        """
        Demultiplex digital data by extracting individual channel values from packed 16-bit format.

        According to the Intan format, digital input/output data is stored with all 16 channels
        encoded bit-by-bit in each 16-bit word. This method extracts the specified digital channels
        from the packed format into separate uint16 arrays of 0 and 1.

        Parameters
        ----------
        raw_digital_data : ndarray
            Raw digital data in packed 16-bit format where each bit represents a different channel.
        channel_ids : list or array
            List of channel identifiers to extract. Each channel_id must correspond to a digital
            input or output channel.
        i_start : int
            Starting sample index for demultiplexing.
        i_stop : int
            Ending sample index for demultiplexing (exclusive).

        Returns
        -------
        ndarray
            Demultiplexed digital data with shape (i_stop-i_start, len(channel_ids)),
            containing 0 or 1 values for each requested channel.

        Notes
        -----
        In the Intan format, digital channels are packed into 16-bit words where each bit position
        corresponds to a specific channel number. For example, with digital inputs 0, 4, and 5
        set high and the rest low, the 16-bit word would be 2^0 + 2^4 + 2^5 = 1 + 16 + 32 = 49.

        The native_order property for each channel corresponds to its bit position in the packed word.

        """
        dtype = np.uint16  # We fix this to match the memmap dtype
        output = np.zeros((i_stop - i_start, len(channel_ids)), dtype=dtype)

        for channel_index, channel_id in enumerate(channel_ids):
            native_order = self.native_channel_order[channel_id]
            mask = 1 << native_order
            demultiplex_data = np.bitwise_and(raw_digital_data, mask) > 0
            output[:, channel_index] = demultiplex_data[i_start:i_stop].flatten()

        return output

    def _decode_current_from_stim_data(self, raw_stim_data, i_start, i_stop):
        """
        Demultiplex stimulation data by extracting current values from packed 16-bit format.

        According to the Intan RHS data format, stimulation current is stored in the lower 9 bits
        of each 16-bit word: 8 bits for magnitude and 1 bit for sign. The upper bits contain
        flags for compliance limit, charge recovery, and amplifier settle.

        Parameters
        ----------
        raw_stim_data : ndarray
            Raw stimulation data in packed 16-bit format.
        i_start : int
            Starting sample index for demultiplexing.
        i_stop : int
            Ending sample index for demultiplexing (exclusive).

        Returns
        -------
        ndarray
            Demultiplexed stimulation current values in amperes, preserving the original
            array dimensions. The output values need to be multiplied by the stim_step_size
            parameter (from header) to obtain the actual current in amperes.

        Notes
        -----
        Bit structure of each 16-bit stimulation word:
        - Bits 0-7: Current magnitude
        - Bit 8: Sign bit (1 = negative current)
        - Bits 9-13: Unused (always zero)
        - Bit 14: Amplifier settle flag (1 = activated)
        - Bit 15: Charge recovery flag (1 = activated)
        - Bit 16 (MSB): Compliance limit flag (1 = limit reached)

        The actual current value in amperes is obtained by multiplying the
        output by the 'stim_step_size' parameter from the file header. These scaled values can be
        obtained with the `rescale_signal_raw_to_float` function.
        """
        # Get the relevant portion of the data
        data = raw_stim_data[i_start:i_stop]

        # Extract current value (bits 0-8)
        magnitude = np.bitwise_and(data, 0xFF)
        sign_bit = np.bitwise_and(np.right_shift(data, 8), 0x01)  # Shift right by 8 bits to get the sign bit

        # Apply sign to current values
        # We need to cast to int16 to handle negative values correctly
        # The max value of 8 bits is 255 so the casting is safe as there are non-negative values
        magnitude = magnitude.astype(np.int16)
        current = np.where(sign_bit == 1, -magnitude, magnitude)

        # Note: If needed, other flag bits could be extracted as follows:
        # compliance_flag = np.bitwise_and(np.right_shift(data, 15), 0x01).astype(bool)  # Bit 16 (MSB)
        # charge_recovery_flag = np.bitwise_and(np.right_shift(data, 14), 0x01).astype(bool)  # Bit 15
        # amp_settle_flag = np.bitwise_and(np.right_shift(data, 13), 0x01).astype(bool)  # Bit 14
        # These could be returned as a structured array or dictionary if needed

        return current

    def get_intan_timestamps(self, i_start=None, i_stop=None):
        """
        Retrieves the sample indices from the Intan raw data within a specified range.

        Note that sample indices are called timestamps in the Intan format but they are
        in fact just sample indices. This function extracts the sample index timestamps
        from Intan files, which represent  relative time points in sample units (not absolute time).
        These indices can be  particularly useful when working with recordings that have discontinuities.

        Parameters
        ----------
        i_start : int, optional
            The starting index from which to retrieve sample indices. If None, starts from 0.
        i_stop : int, optional
            The stopping index up to which to retrieve sample indices (exclusive).
            If None, retrieves all available indices from i_start onward.

        Returns
        -------
        timestamps : ndarray
            The flattened array of sample indices within the specified range.

        Notes
        -----
        - Sample indices can be converted to seconds by dividing by the sampling rate of the amplifier stream.
        - The function automatically handles different file formats:
        * header-attached: Timestamps are extracted directly from the timestamp field
        * one-file-per-signal: Timestamps are read from the timestamp stream
        * one-file-per-channel: Timestamps are read from the first channel in the timestamp stream
        - When recordings have discontinuities (indicated by the `discontinuous_timestamps`
        attribute being True), these indices allow for proper temporal alignment of the data.
        """
        if i_start is None:
            i_start = 0

        # Get the timestamps based on file format
        if self.file_format == "header-attached":
            timestamps = self._raw_data["timestamp"]
        elif self.file_format == "one-file-per-signal":
            timestamps = self._raw_data["timestamp"]
        elif self.file_format == "one-file-per-channel":
            timestamps = self._raw_data["timestamp"][0]

        # TODO if possible ensure that timestamps memmaps are always of correct shape to avoid memory copy here.
        timestamps = timestamps.flatten() if timestamps.ndim > 1 else timestamps

        if i_stop is None:
            return timestamps[i_start:]
        else:
            return timestamps[i_start:i_stop]

    def _assert_timestamp_continuity(self):
        """
        Asserts the continuity of timestamps in the data.

        This method verifies that the timestamps in the raw data are sequential,
        indicating a continuous recording. If discontinuities are found, a flag
        is set to indicate potential data integrity issues, and an error is raised
        unless `ignore_integrity_checks` is True.

        Raises
        ------
        NeoReadWriteError
            If timestamps are not continuous and `ignore_integrity_checks` is False.
            The error message includes a table detailing the discontinuities found.
        """
        # check timestamp continuity
        timestamps = self.get_intan_timestamps()

        discontinuous_timestamps = np.diff(timestamps) != 1
        timestamps_are_not_contiguous = np.any(discontinuous_timestamps)
        if timestamps_are_not_contiguous:
            # Mark a flag that can be checked after parsing the header to see if the timestamps are continuous or not
            self.discontinuous_timestamps = True
            if not self.ignore_integrity_checks:
                error_msg = (
                    "\nTimestamps are not continuous, likely due to a corrupted file or inappropriate file merge.\n"
                    "To open the file anyway, initialize the reader with `ignore_integrity_checks=True`.\n\n"
                    "Discontinuities Found:\n"
                    "+-----------------+-----------------+-----------------+-----------------------+\n"
                    "| Discontinuity   | Previous        | Next            | Time Difference       |\n"
                    "| Index           | (Frames)        | (Frames)        | (Seconds)             |\n"
                    "+-----------------+-----------------+-----------------+-----------------------+\n"
                )

                amplifier_sampling_rate = self._global_info["sampling_rate"]
                for discontinuity_index in np.where(discontinuous_timestamps)[0]:
                    prev_ts = timestamps[discontinuity_index]
                    next_ts = timestamps[discontinuity_index + 1]
                    time_diff = (next_ts - prev_ts) / amplifier_sampling_rate

                    error_msg += (
                        f"| {discontinuity_index + 1:>15,} | {prev_ts:>15,} | {next_ts:>15,} | {time_diff:>21.6f} |\n"
                    )

                error_msg += "+-----------------+-----------------+-----------------+-----------------------+\n"

                raise NeoReadWriteError(error_msg)


###################################
# Header reading helper functions


def read_qstring(f):
    """
    Reads the optional notes included in the Intan RHX software

    Parameters
    ----------
    f: BinaryIO
        The file object

    Returns
    -------
    txt: str
        The string
    """
    length = np.fromfile(f, dtype="uint32", count=1)[0]
    if length == 0xFFFFFFFF or length == 0:
        return ""
    txt = f.read(length).decode("utf-16")
    return txt


def read_variable_header(f, header: list):
    """
    Reads the information from the binary file for the header info dict

    Parameters
    ----------
    f: BinaryIO
        The file object
    header: list[tuple]
        The list of header sections along with their associated dtype

    Returns
    -------
    info: dict
        The dictionary containing the information contained in the header
    """
    info = {}
    for field_name, field_type in header:
        if field_type == "QString":
            field_value = read_qstring(f)
        else:
            field_value = np.fromfile(f, dtype=field_type, count=1)[0]
        info[field_name] = field_value
    return info


###############
# RHS ZONE

rhs_global_header = [
    ("magic_number", "uint32"),  # 0xD69127AC
    ("major_version", "int16"),
    ("minor_version", "int16"),
    ("sampling_rate", "float32"),
    ("dsp_enabled", "int16"),
    ("actual_dsp_cutoff_frequency", "float32"),
    ("actual_lower_bandwidth", "float32"),
    ("actual_lower_settle_bandwidth", "float32"),
    ("actual_upper_bandwidth", "float32"),
    ("desired_dsp_cutoff_frequency", "float32"),
    ("desired_lower_bandwidth", "float32"),
    ("desired_lower_settle_bandwidth", "float32"),
    ("desired_upper_bandwidth", "float32"),
    ("notch_filter_mode", "int16"),
    ("desired_impedance_test_frequency", "float32"),
    ("actual_impedance_test_frequency", "float32"),
    ("amp_settle_mode", "int16"),
    ("charge_recovery_mode", "int16"),
    ("stim_step_size", "float32"),
    ("recovery_current_limit", "float32"),
    ("recovery_target_voltage", "float32"),
    ("note1", "QString"),
    ("note2", "QString"),
    ("note3", "QString"),
    ("dc_amplifier_data_saved", "int16"),
    ("board_mode", "int16"),
    ("ref_channel_name", "QString"),
    ("nb_signal_group", "int16"),
]

rhs_signal_group_header = [
    ("signal_group_name", "QString"),
    ("signal_group_prefix", "QString"),
    ("signal_group_enabled", "int16"),
    ("channel_num", "int16"),
    ("amplified_channel_num", "int16"),
]

rhs_signal_channel_header = [
    ("native_channel_name", "QString"),
    ("custom_channel_name", "QString"),
    ("native_order", "int16"),
    ("custom_order", "int16"),
    ("signal_type", "int16"),
    ("channel_enabled", "int16"),
    ("chip_channel_num", "int16"),
    ("command_stream", "int16"),
    ("board_stream_num", "int16"),
    ("spike_scope_trigger_mode", "int16"),
    ("spike_scope_voltage_thresh", "int16"),
    ("spike_scope_digital_trigger_channel", "int16"),
    ("spike_scope_digital_edge_polarity", "int16"),
    ("electrode_impedance_magnitude", "float32"),
    ("electrode_impedance_phase", "float32"),
]

stream_id_to_stream_name_rhs = {
    "0": "RHS2000 amplifier channel",
    "3": "USB board ADC input channel",
    "4": "USB board ADC output channel",
    "5": "USB board digital input channel",
    "6": "USB board digital output channel",
    "10": "DC Amplifier channel",
    "11": "Stim channel",
}


def read_rhs(filename, file_format: str):
    BLOCK_SIZE = 128  # sample per block

    with open(filename, mode="rb") as f:
        global_info = read_variable_header(f, rhs_global_header)

        # We use signal_type in the header as stream_id in neo with the following
        # complications: for rhs files in the header-attaached stream_id 0 also
        # contains information for stream_id 10 and stream_id 11 so we need to
        # break these up. See notes throughout code base; for timestamps we always
        # force them to be the last stream_id.
        stream_names = stream_id_to_stream_name_rhs.values()
        stream_name_to_channel_info_list = {name: [] for name in stream_names}
        if not file_format == "header-attached":
            memmap_data_dtype = {name: [] for name in stream_names}
        for g in range(global_info["nb_signal_group"]):
            group_info = read_variable_header(f, rhs_signal_group_header)

            if bool(group_info["signal_group_enabled"]):
                for c in range(group_info["channel_num"]):
                    chan_info = read_variable_header(f, rhs_signal_channel_header)
                    if chan_info["signal_type"] in (1, 2):
                        error_msg = (
                            "signal_type of 1 or 2 does not exist for rhs files. If you have an rhs file "
                            "with these formats open an issue on the python-neo github page"
                        )
                        raise NeoReadWriteError(error_msg)
                    if bool(chan_info["channel_enabled"]):
                        stream_id = str(chan_info["signal_type"])
                        stream_name = stream_id_to_stream_name_rhs[stream_id]
                        stream_name_to_channel_info_list[stream_name].append(chan_info)

        # Build a dictionary with channel count
        special_cases_for_counting = [
            "DC Amplifier channel",
            "Stim channel",
        ]
        names_to_count = [name for name in stream_names if name not in special_cases_for_counting]
        channel_number_dict = {name: len(stream_name_to_channel_info_list[name]) for name in names_to_count}

        # Each DC amplifier channel has a corresponding RHS2000 amplifier channel
        channel_number_dict["DC Amplifier channel"] = channel_number_dict["RHS2000 amplifier channel"]

        if file_format == "one-file-per-channel":
            # There is a way to shut off saving amplifier data and only keeping the DC amplifier or shutting off all amplifier file saving,
            # so we need to count the number of files we find instead of relying on the header.
            raw_file_paths_dict = create_one_file_per_channel_dict_rhs(dirname=filename.parent)
            channel_number_dict["Stim channel"] = len(raw_file_paths_dict["Stim channel"])
            # Moreover, even if the amplifier channels are in the header their files are dropped
            channel_number_dict["RHS2000 amplifier channel"] = len(raw_file_paths_dict["RHS2000 amplifier channel"])
        else:
            channel_number_dict["Stim channel"] = channel_number_dict["RHS2000 amplifier channel"]

        header_size = f.tell()

    sr = global_info["sampling_rate"]

    # construct dtype by re-ordering channels by types
    ordered_channel_info = []
    if file_format == "header-attached":
        memmap_data_dtype = [("timestamp", "int32", BLOCK_SIZE)]
    else:
        memmap_data_dtype["timestamp"] = "int32"
        channel_number_dict["timestamp"] = 1

    if channel_number_dict["RHS2000 amplifier channel"] > 0:
        for chan_info in stream_name_to_channel_info_list["RHS2000 amplifier channel"]:
            chan_info["sampling_rate"] = sr
            chan_info["units"] = "uV"
            chan_info["gain"] = 0.195
            if file_format == "header-attached":
                chan_info["offset"] = -32768 * 0.195
            else:
                chan_info["offset"] = 0.0
            if file_format == "header-attached":
                chan_info["dtype"] = "uint16"
            else:
                chan_info["dtype"] = "int16"
            ordered_channel_info.append(chan_info)
            if file_format == "header-attached":
                name = chan_info["native_channel_name"]
                memmap_data_dtype += [(name, "uint16", BLOCK_SIZE)]
            else:
                memmap_data_dtype["RHS2000 amplifier channel"] = "int16"

    if bool(global_info["dc_amplifier_data_saved"]):
        # if we have dc amp we need to grab the correct number of channels
        # the channel number is the same as the count for amplifier data
        for chan_info in stream_name_to_channel_info_list["RHS2000 amplifier channel"]:
            chan_info_dc = dict(chan_info)
            name = chan_info["native_channel_name"]
            chan_info_dc["native_channel_name"] = name + "_DC"
            chan_info_dc["custom_channel_name"] = chan_info_dc["custom_channel_name"] + "_DC"
            chan_info_dc["sampling_rate"] = sr
            chan_info_dc["units"] = "mV"
            chan_info_dc["gain"] = 19.23
            chan_info_dc["offset"] = -512 * 19.23
            chan_info_dc["signal_type"] = 10  # put it in another group
            chan_info_dc["dtype"] = "uint16"
            ordered_channel_info.append(chan_info_dc)
            if file_format == "header-attached":
                memmap_data_dtype += [(name + "_DC", "uint16", BLOCK_SIZE)]
            else:
                memmap_data_dtype["DC Amplifier channel"] = "uint16"

    # I can't seem to get stim files to generate for one-file-per-channel
    # so ideally at some point we need test data to confirm this is true
    # based on what Heberto and I read in the docs

    # Add stim channels
    for chan_info in stream_name_to_channel_info_list["RHS2000 amplifier channel"]:
        # stim channel presence is not indicated in the header so for some formats each amplifier channel has a stim channel, but for other formats this isn't the case.
        if file_format == "one-file-per-channel":
            # Some amplifier channels don't have a corresponding stim channel,
            # so we need to make sure we don't add channel info for stim channels that don't exist.
            # In this case, if the stim channel has no data, there won't be a file for it.
            stim_file_paths = raw_file_paths_dict["Stim channel"]
            amplifier_native_name = chan_info["native_channel_name"]
            stim_file_exists = any([amplifier_native_name in stim_file.stem for stim_file in stim_file_paths])
            if not stim_file_exists:
                continue

        chan_info_stim = dict(chan_info)
        name = chan_info["native_channel_name"]
        chan_info_stim["native_channel_name"] = name + "_STIM"
        chan_info_stim["custom_channel_name"] = chan_info_stim["custom_channel_name"] + "_STIM"
        chan_info_stim["sampling_rate"] = sr
        chan_info_stim["units"] = "A"  # Amps
        chan_info_stim["gain"] = global_info["stim_step_size"]
        chan_info_stim["offset"] = 0.0
        chan_info_stim["signal_type"] = 11  # put it in another group
        chan_info_stim["dtype"] = "int16"  # this change is due to bit decoding see note below
        ordered_channel_info.append(chan_info_stim)
        # Note that the data on disk is uint16 but the data is
        # then decoded as int16 so the chan_info is int16
        if file_format == "header-attached":
            memmap_data_dtype += [(name + "_STIM", "uint16", BLOCK_SIZE)]
        else:
            memmap_data_dtype["Stim channel"] = "uint16"

    # No supply or aux for rhs files (ie no stream_id 1 and 2)
    # We have an error above that requests test files to help if the spec is changed
    # in the future.

    for stream_name in ["USB board ADC input channel", "USB board ADC output channel"]:
        for chan_info in stream_name_to_channel_info_list[stream_name]:
            chan_info["sampling_rate"] = sr
            chan_info["units"] = "V"
            chan_info["gain"] = 0.0003125
            chan_info["offset"] = -32768 * 0.0003125
            chan_info["dtype"] = "uint16"
            ordered_channel_info.append(chan_info)
            if file_format == "header-attached":
                name = chan_info["native_channel_name"]
                memmap_data_dtype += [(name, "uint16", BLOCK_SIZE)]
            else:
                memmap_data_dtype[stream_name] = "uint16"

    for stream_name in ["USB board digital input channel", "USB board digital output channel"]:
        for chan_info in stream_name_to_channel_info_list[stream_name]:
            chan_info["sampling_rate"] = sr
            # arbitrary units are used to indicate that Intan does not
            # store raw voltages but only the boolean TTL state
            chan_info["units"] = "a.u."
            chan_info["gain"] = 1.0
            chan_info["offset"] = 0.0
            chan_info["dtype"] = "uint16"
            ordered_channel_info.append(chan_info)

        # Note that all the channels are packed in one buffer, so the data type only needs to be added once
        if len(stream_name_to_channel_info_list[stream_name]) > 0:
            if file_format == "header-attached":
                memmap_data_dtype += [(stream_name, "uint16", BLOCK_SIZE)]
            elif file_format == "one-file-per-channel":
                memmap_data_dtype[stream_name] = "uint16"
            elif file_format == "one-file-per-signal":
                memmap_data_dtype[stream_name] = "uint16"

    # per discussion with Intan developers before version 3 of their software the 'notch_filter_mode'
    # was a request for postprocessing to be done in one of their scripts. From version 3+ the notch
    # filter is now applied to the data in realtime and only the post notched amplifier data is
    # saved.
    if global_info["notch_filter_mode"] == 2 and global_info["major_version"] >= 3:
        global_info["notch_filter"] = "60Hz"
    elif global_info["notch_filter_mode"] == 1 and global_info["major_version"] >= 3:
        global_info["notch_filter"] = "50Hz"
    else:
        global_info["notch_filter"] = False

    if not file_format == "header-attached":
        # filter out dtypes without any values
        memmap_data_dtype = {k: v for (k, v) in memmap_data_dtype.items() if len(v) > 0}
        channel_number_dict = {k: v for (k, v) in channel_number_dict.items() if v > 0}

    return global_info, ordered_channel_info, memmap_data_dtype, header_size, BLOCK_SIZE, channel_number_dict


###############
# RHD ZONE

rhd_global_header_base = [
    ("magic_number", "uint32"),  # 0xC6912702
    ("major_version", "int16"),
    ("minor_version", "int16"),
]

rhd_global_header_part1 = [
    ("sampling_rate", "float32"),
    ("dsp_enabled", "int16"),
    ("actual_dsp_cutoff_frequency", "float32"),
    ("actual_lower_bandwidth", "float32"),
    ("actual_upper_bandwidth", "float32"),
    ("desired_dsp_cutoff_frequency", "float32"),
    ("desired_lower_bandwidth", "float32"),
    ("desired_upper_bandwidth", "float32"),
    ("notch_filter_mode", "int16"),
    ("desired_impedance_test_frequency", "float32"),
    ("actual_impedance_test_frequency", "float32"),
    ("note1", "QString"),
    ("note2", "QString"),
    ("note3", "QString"),
]

rhd_global_header_v11 = [
    ("num_temp_sensor_channels", "int16"),
]

rhd_global_header_v13 = [
    ("eval_board_mode", "int16"),
]

rhd_global_header_v20 = [
    ("reference_channel", "QString"),
]

rhd_global_header_final = [
    ("nb_signal_group", "int16"),
]

rhd_signal_group_header = [
    ("signal_group_name", "QString"),
    ("signal_group_prefix", "QString"),
    ("signal_group_enabled", "int16"),
    ("channel_num", "int16"),
    ("amplified_channel_num", "int16"),
]

rhd_signal_channel_header = [
    ("native_channel_name", "QString"),
    ("custom_channel_name", "QString"),
    ("native_order", "int16"),
    ("custom_order", "int16"),
    ("signal_type", "int16"),
    ("channel_enabled", "int16"),
    ("chip_channel_num", "int16"),
    ("board_stream_num", "int16"),
    ("spike_scope_trigger_mode", "int16"),
    ("spike_scope_voltage_thresh", "int16"),
    ("spike_scope_digital_trigger_channel", "int16"),
    ("spike_scope_digital_edge_polarity", "int16"),
    ("electrode_impedance_magnitude", "float32"),
    ("electrode_impedance_phase", "float32"),
]

stream_id_to_stream_name_rhd = {
    "0": "RHD2000 amplifier channel",
    "1": "RHD2000 auxiliary input channel",
    "2": "RHD2000 supply voltage channel",
    "3": "USB board ADC input channel",
    "4": "USB board digital input channel",
    "5": "USB board digital output channel",
}


def read_rhd(filename, file_format: str):
    """Function for reading the rhd file header

    Parameters
    ----------
    filename: str | Path
        The filename of the *.rhd file
    file_format: 'header-attached' | 'one-file-per-signal' | 'one-file-per-channel'
        Whether the header is included with the rest of the data ('header-attached')
        Or as a standalone file ('one-file-per-signal' or 'one-file-per-channel')
    """
    with open(filename, mode="rb") as f:

        global_info = read_variable_header(f, rhd_global_header_base)

        version = Version(f"{global_info['major_version']}.{global_info['minor_version']}")

        # the header size depends on the version :-(
        header = list(rhd_global_header_part1)  # make a copy

        if version >= Version("1.1"):
            header = header + rhd_global_header_v11
        else:
            global_info["num_temp_sensor_channels"] = 0

        if version >= Version("1.3"):
            header = header + rhd_global_header_v13
        else:
            global_info["eval_board_mode"] = 0

        if version >= Version("2.0"):
            header = header + rhd_global_header_v20
        else:
            global_info["reference_channel"] = ""

        header = header + rhd_global_header_final

        global_info.update(read_variable_header(f, header))

        # read channel group and channel header
        stream_names = stream_id_to_stream_name_rhd.values()
        stream_name_to_channel_info_list = {name: [] for name in stream_names}
        if not file_format == "header-attached":
            memmap_data_dtype = {name: [] for name in stream_names}
            memmap_data_dtype["timestamp"] = []
        for g in range(global_info["nb_signal_group"]):
            group_info = read_variable_header(f, rhd_signal_group_header)

            if bool(group_info["signal_group_enabled"]):
                for c in range(group_info["channel_num"]):
                    chan_info = read_variable_header(f, rhd_signal_channel_header)
                    if bool(chan_info["channel_enabled"]):
                        stream_id = str(chan_info["signal_type"])
                        stream_name = stream_id_to_stream_name_rhd[stream_id]
                        stream_name_to_channel_info_list[stream_name].append(chan_info)

            channel_number_dict = {name: len(stream_name_to_channel_info_list[name]) for name in stream_names}

        header_size = f.tell()

    sr = global_info["sampling_rate"]

    # construct the data block dtype and reorder channels
    if version >= Version("2.0"):
        BLOCK_SIZE = 128
    else:
        BLOCK_SIZE = 60  # 256 channels

    ordered_channel_info = []

    if version >= Version("1.2"):
        if file_format == "header-attached":
            memmap_data_dtype = [("timestamp", "int32", BLOCK_SIZE)]
        else:
            memmap_data_dtype["timestamp"] = "int32"
            channel_number_dict["timestamp"] = 1
    else:
        if file_format == "header-attached":
            memmap_data_dtype = [("timestamp", "uint32", BLOCK_SIZE)]
        else:
            memmap_data_dtype["timestamp"] = "uint32"
            channel_number_dict["timestamp"] = 1

    for chan_info in stream_name_to_channel_info_list["RHD2000 amplifier channel"]:
        chan_info["sampling_rate"] = sr
        chan_info["units"] = "uV"
        chan_info["gain"] = 0.195
        if file_format == "header-attached":
            chan_info["offset"] = -32768 * 0.195
            chan_info["dtype"] = "uint16"
        else:
            chan_info["offset"] = 0.0
            chan_info["dtype"] = "int16"
        ordered_channel_info.append(chan_info)

        if file_format == "header-attached":
            name = chan_info["native_channel_name"]
            memmap_data_dtype += [(name, "uint16", BLOCK_SIZE)]
        else:
            memmap_data_dtype["RHD2000 amplifier channel"] = "int16"

    for chan_info in stream_name_to_channel_info_list["RHD2000 auxiliary input channel"]:
        chan_info["sampling_rate"] = sr / 4.0
        chan_info["units"] = "V"
        chan_info["gain"] = 0.0000374
        chan_info["offset"] = 0.0
        chan_info["dtype"] = "uint16"
        ordered_channel_info.append(chan_info)
        if file_format == "header-attached":
            name = chan_info["native_channel_name"]
            memmap_data_dtype += [(name, "uint16", BLOCK_SIZE // 4)]
        else:
            memmap_data_dtype["RHD2000 auxiliary input channel"] = "uint16"

    for chan_info in stream_name_to_channel_info_list["RHD2000 supply voltage channel"]:
        chan_info["sampling_rate"] = sr / BLOCK_SIZE
        chan_info["units"] = "V"
        chan_info["gain"] = 0.0000748
        chan_info["offset"] = 0.0
        chan_info["dtype"] = "uint16"
        ordered_channel_info.append(chan_info)
        if file_format == "header-attached":
            name = chan_info["native_channel_name"]
            memmap_data_dtype += [(name, "uint16")]
        else:
            memmap_data_dtype["RHD2000 supply voltage channel"] = "uint16"

    # temperature is not an official channel in the header
    for i in range(global_info["num_temp_sensor_channels"]):
        name = f"temperature_{i}"
        chan_info = {"native_channel_name": name, "signal_type": 20}
        chan_info["sampling_rate"] = sr / BLOCK_SIZE
        chan_info["units"] = "Celsius"
        chan_info["gain"] = 0.001
        chan_info["offset"] = 0.0
        chan_info["dtype"] = "int16"
        ordered_channel_info.append(chan_info)
        memmap_data_dtype += [(name, "int16")]

    # 3: USB board ADC input channel
    for chan_info in stream_name_to_channel_info_list["USB board ADC input channel"]:
        chan_info["sampling_rate"] = sr
        chan_info["units"] = "V"
        if global_info["eval_board_mode"] == 0:
            chan_info["gain"] = 0.000050354
            chan_info["offset"] = 0.0
        elif global_info["eval_board_mode"] == 1:
            chan_info["gain"] = 0.00015259
            chan_info["offset"] = -32768 * 0.00015259
        elif global_info["eval_board_mode"] == 13:
            chan_info["gain"] = 0.0003125
            chan_info["offset"] = -32768 * 0.0003125
        chan_info["dtype"] = "uint16"
        ordered_channel_info.append(chan_info)
        if file_format == "header-attached":
            name = chan_info["native_channel_name"]
            memmap_data_dtype += [(name, "uint16", BLOCK_SIZE)]
        else:
            memmap_data_dtype["USB board ADC input channel"] = "uint16"

    for stream_name in ["USB board digital input channel", "USB board digital output channel"]:
        for chan_info in stream_name_to_channel_info_list[stream_name]:
            chan_info["sampling_rate"] = sr
            # arbitrary units are used to indicate that Intan does not
            # store raw voltages but only the boolean TTL state
            chan_info["units"] = "a.u."
            chan_info["gain"] = 1.0
            chan_info["offset"] = 0.0
            chan_info["dtype"] = "uint16"
            ordered_channel_info.append(chan_info)

        # Note that all the channels are packed in one buffer, so the data type only needs to be added once
        if len(stream_name_to_channel_info_list[stream_name]) > 0:
            if file_format == "header-attached":
                memmap_data_dtype += [(stream_name, "uint16", BLOCK_SIZE)]
            elif file_format == "one-file-per-channel":
                memmap_data_dtype[stream_name] = "uint16"
            elif file_format == "one-file-per-signal":
                memmap_data_dtype[stream_name] = "uint16"

    # per discussion with Intan developers before version 3 of their software the 'notch_filter_mode'
    # was a request for postprocessing to be done in one of their scripts. From version 3+ the notch
    # filter is now applied to the data in realtime and only the post notched amplifier data is
    # saved.
    if global_info["notch_filter_mode"] == 2 and version >= Version("3.0"):
        global_info["notch_filter"] = "60Hz"
    elif global_info["notch_filter_mode"] == 1 and version >= Version("3.0"):
        global_info["notch_filter"] = "50Hz"
    else:
        global_info["notch_filter"] = False

    if not file_format == "header-attached":
        # filter out dtypes without any values
        memmap_data_dtype = {k: v for (k, v) in memmap_data_dtype.items() if len(v) > 0}
        channel_number_dict = {k: v for (k, v) in channel_number_dict.items() if v > 0}

    return global_info, ordered_channel_info, memmap_data_dtype, header_size, BLOCK_SIZE, channel_number_dict


##########################################################################
# RHX Zone for Binary Files (note this is for version 3+ of Intan software)
# This section provides all possible headerless binary files in both the rhs and rhd
# formats.

digital_stream_names = ["USB board digital input channel", "USB board digital output channel"]


stream_name_to_file_name_rhd = {
    "RHD2000 amplifier channel": "amplifier.dat",
    "RHD2000 auxiliary input channel": "auxiliary.dat",
    "RHD2000 supply voltage channel": "supply.dat",
    "USB board ADC input channel": "analogin.dat",
    "USB board digital input channel": "digitalin.dat",
    "USB board digital output channel": "digitalout.dat",
}


def create_one_file_per_signal_dict_rhd(dirname):
    """Function for One File Per Signal Type

    Parameters
    ----------
    dirname: pathlib.Path
        The folder to explore

    Returns
    -------
    raw_files_paths_dict: dict
        A dict of all the file paths
    """

    raw_file_paths_dict = {}
    for stream_name, file_name in stream_name_to_file_name_rhd.items():
        if Path(dirname / file_name).is_file():
            raw_file_paths_dict[stream_name] = Path(dirname / file_name)

    raw_file_paths_dict["timestamp"] = Path(dirname / "time.dat")

    return raw_file_paths_dict


stream_name_to_file_name_rhs = {
    "RHS2000 amplifier channel": "amplifier.dat",
    "RHS2000 auxiliary input channel": "auxiliary.dat",
    "RHS2000 supply voltage channel": "supply.dat",
    "USB board ADC input channel": "analogin.dat",
    "USB board ADC output channel": "analogout.dat",
    "USB board digital input channel": "digitalin.dat",
    "USB board digital output channel": "digitalout.dat",
}


def create_one_file_per_signal_dict_rhs(dirname):
    """Function for One File Per Signal Type

    Parameters
    ----------
    dirname: pathlib.Path
        The folder to explore

    Returns
    -------
    raw_files_paths_dict: dict
        A dict of all the file paths
    """

    raw_file_paths_dict = {}
    for stream_name, file_name in stream_name_to_file_name_rhs.items():
        if Path(dirname / file_name).is_file():
            raw_file_paths_dict[stream_name] = Path(dirname / file_name)

    raw_file_paths_dict["timestamp"] = Path(dirname / "time.dat")
    if Path(dirname / "dcamplifier.dat").is_file():
        raw_file_paths_dict["DC Amplifier channel"] = Path(dirname / "dcamplifier.dat")
    if Path(dirname / "stim.dat").is_file():
        raw_file_paths_dict["Stim channel"] = Path(dirname / "stim.dat")

    return raw_file_paths_dict


# RHD Binary Files for One File Per Channel
stream_name_to_file_prefix_rhd = {
    "RHD2000 amplifier channel": "amp",
    "RHD2000 auxiliary input channel": "aux",
    "RHD2000 supply voltage channel": "vdd",
    "USB board ADC input channel": "board-ANALOG-IN",
    "USB board digital input channel": "board-DIGITAL-IN",
    "USB board digital output channel": "board-DIGITAL-OUT",
}


def create_one_file_per_channel_dict_rhd(dirname):
    """Utility function for One File Per Channel

    Parameters
    ----------
    dirname: pathlib.Path
        The folder to explore

    Returns
    -------
    raw_files_paths_dict: dict
        A dict of all the file paths
    """
    # Developer note, at the moment, the channels require to be n in order
    # See https://github.com/NeuralEnsemble/python-neo/issues/1599 fo

    file_names = dirname.glob("**/*.dat")
    files = [file for file in file_names if file.is_file()]
    raw_file_paths_dict = {}
    for stream_name, file_prefix in stream_name_to_file_prefix_rhd.items():
        stream_files = [file for file in files if file_prefix in file.name]
        # Map file name to channel name amp-A-000.dat -> A-000  amp-B-006.dat -> B-006  etc
        file_path_to_channel_name = lambda x: "-".join(x.stem.split("-")[1:])
        sorted_stream_files = sorted(stream_files, key=file_path_to_channel_name)
        raw_file_paths_dict[stream_name] = sorted_stream_files

    raw_file_paths_dict["timestamp"] = [Path(dirname / "time.dat")]
    return raw_file_paths_dict


# RHS Binary Files for One File Per Channel
stream_name_to_file_prefix_rhs = {
    "RHS2000 amplifier channel": "amp",
    "RHS2000 auxiliary input channel": "aux",
    "RHS2000 supply voltage channel": "vdd",
    "USB board ADC input channel": "board-ANALOG-IN",
    "USB board ADC output channel": "board-ANALOG-OUT",
    "USB board digital input channel": "board-DIGITAL-IN",
    "USB board digital output channel": "board-DIGITAL-OUT",
}


def create_one_file_per_channel_dict_rhs(
    dirname,
):
    """Utility function for One File Per Channel

    Parameters
    ----------
    dirname: pathlib.Path
        The folder to explore

    Returns
    -------
    raw_files_paths_dict: dict
        A dict of all the file paths
    """

    file_names = dirname.glob("**/*.dat")
    files = [file for file in file_names if file.is_file()]
    raw_file_paths_dict = {}
    for stream_name, file_prefix in stream_name_to_file_prefix_rhs.items():
        stream_files = [file for file in files if file_prefix in file.name]
        # Map file name to channel name amp-A-000.dat -> A-000  amp-B-006.dat -> B-006  etc
        file_path_to_channel_name = lambda x: "-".join(x.stem.split("-")[1:])
        sorted_stream_files = sorted(stream_files, key=file_path_to_channel_name)
        raw_file_paths_dict[stream_name] = sorted_stream_files

    raw_file_paths_dict["timestamp"] = [Path(dirname / "time.dat")]
    raw_file_paths_dict["DC Amplifier channel"] = [file for file in files if "dc-" in file.name]
    # we can find the files, but I can see how to read them out of header
    # so for now we don't expose the stim files in one-file-per-channel
    raw_file_paths_dict["Stim channel"] = [file for file in files if "stim-" in file.name]

    return raw_file_paths_dict



"""
IgorProIO Demo (BROKEN)
=======================

"""

###########################################################
# Import our packages
import os
from urllib.request import urlretrieve
import zipfile
import matplotlib.pyplot as plt
from neo.io import get_io

#############################################################
# Then download some data
# Downloaded from Human Brain Project Collaboratory
# Digital Reconstruction of Neocortical Microcircuitry (nmc-portal)
# http://microcircuits.epfl.ch/#/animal/8ecde7d1-b2d2-11e4-b949-6003088da632
# NOTE: this dataset is not found as the link is broken.

# datafile_url = "https://microcircuits.epfl.ch/data/released_data/B95.zip"
# filename_zip = "B95.zip"
# filename = "grouped_ephys/B95/B95_Ch0_IDRest_107.ibw"
# urlretrieve(datafile_url, filename_zip)

# zip_ref = zipfile.ZipFile(filename_zip)  # create zipfile object
# zip_ref.extract(path=".", member=filename)  # extract file to dir
# zip_ref.close()

# ######################################################
# # Once we have our data we can use `get_io` to find an
# # io (Igor in this case). Then we read the analogsignals
# # Finally we will make some nice plots
# reader = get_io(filename)
# signal = reader.read_analogsignal()
# plt.plot(signal.times, signal)
# plt.xlabel(signal.sampling_period.dimensionality)
# plt.ylabel(signal.dimensionality)

# plt.show()
